from csle_common.dao.management.management_user import ManagementUser
from csle_common.dao.management.session_token import SessionToken


class TestManagementDaoSuite:
    """
    Test suite for management data access objects (DAOs)
    """

    def test_management_user(self, example_management_user: ManagementUser) -> None:
        """
        Tests creation and dict conversion of the ManagementUser DAO

        :param example_management_user: an example ManagementUser
        :return: None
        """
        assert isinstance(example_management_user.to_dict(), dict)
        assert isinstance(ManagementUser.from_dict(example_management_user.to_dict()), ManagementUser)
        assert ManagementUser.from_dict(example_management_user.to_dict()).to_dict() == \
               example_management_user.to_dict()
        assert ManagementUser.from_dict(example_management_user.to_dict()) == example_management_user

    def test_session_token(self, example_session_token: SessionToken) -> None:
        """
        Tests creation and dict conversion of the SessionToken DAO

        :param example_session_token: an example SessionToken
        :return: None
        """
        assert isinstance(example_session_token.to_dict(), dict)
        assert isinstance(SessionToken.from_dict(example_session_token.to_dict()), SessionToken)
        assert SessionToken.from_dict(example_session_token.to_dict()).to_dict() == example_session_token.to_dict()
        assert SessionToken.from_dict(example_session_token.to_dict()) == example_session_token



from unittest.mock import patch, MagicMock
from csle_common.dao.simulation_config.simulation_env_config import SimulationEnvConfig
from csle_common.controllers.simulation_env_controller import SimulationEnvController


class TestSimulationEnvControllerSuite:
    """
    Test suite for SimulationEnvController
    """

    @patch("csle_common.metastore.metastore_facade.MetastoreFacade.install_simulation")
    def test_install_simulation(self, mock_install_simulation) -> None:
        """
        Test method that installs the simulation configuration in the metastore

        :param mock_install_simulation: mock_install_simulation
        :return: None
        """
        config = MagicMock(spec=SimulationEnvConfig)
        SimulationEnvController.install_simulation(config)
        mock_install_simulation.assert_called_once_with(config=config)

    @patch("csle_common.metastore.metastore_facade.MetastoreFacade.uninstall_simulation")
    def test_uninstall_simulation(self, mock_uninstall_simulation) -> None:
        """
        Test method that uninstalls a simulation config from the metastore

        :param mock_uninstall_simulation: mock_uninstall_simulation
        :return: None
        """
        config = MagicMock(spec=SimulationEnvConfig)
        SimulationEnvController.uninstall_simulation(config)
        mock_uninstall_simulation.assert_called_once_with(config=config)

    @patch("csle_common.metastore.metastore_facade.MetastoreFacade.save_simulation_image")
    def test_save_simulation_image(self, mock_save_simulation_image) -> None:
        """
        Test method that saves the simulation image

        :param mock_save_simulation_image: mock_save_simulation_image
        :return: None
        """
        img = b"image"
        simulation = "test_simulation"
        SimulationEnvController.save_simulation_image(img, simulation)
        mock_save_simulation_image.assert_called_once_with(img=img, simulation_name=simulation)



from typing import Dict, Any
from csle_common.dao.emulation_observation.common.emulation_vulnerability_observation_state \
    import EmulationVulnerabilityObservationState
from csle_common.dao.emulation_config.transport_protocol import TransportProtocol
from csle_base.json_serializable import JSONSerializable


class NiktoVuln(JSONSerializable):
    """
    DTO representing a vulnerability found with a Nikto scan
    """

    def __init__(self, id: str, osvdb_id: int, method: str, iplink: str,
                 namelink: str, uri: str, description: str):
        """
        Initializes the object

        :param id: the id of the vuln
        :param osvdb_id: the osvdb_id of the vuln
        :param method: the method of the vuln
        :param iplink: the iplink of the vuln
        :param namelink: the namelink of the vuln
        :param uri: the uri of the vuln
        :param description: the description of the vuln
        """
        self.id = id
        self.osvdb_id = osvdb_id
        self.method = method
        self.iplink = iplink
        self.namelink = namelink
        self.uri = uri
        self.description = description

    def to_obs(self) -> EmulationVulnerabilityObservationState:
        """
        Converts the object into a VulnerabilityObservationState

        :return: the created VulnerabilityObservationState object
        """
        vuln = EmulationVulnerabilityObservationState(name="nikto_" + str(self.osvdb_id), port=-1,
                                                      protocol=TransportProtocol.TCP,
                                                      cvss=0, osvdbid=self.osvdb_id, description=self.description,
                                                      service="http", credentials=[])
        return vuln

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"id:{self.id}, osvdb_id:{self.osvdb_id}, method:{self.method}, iplink:{self.iplink}, " \
               f"namelink:{self.namelink}, uri:{self.uri}, descr:{self.description}"

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["id"] = self.id
        d["osvdb_id"] = self.osvdb_id
        d["method"] = self.method
        d["iplink"] = self.iplink
        d["namelink"] = self.namelink
        d["uri"] = self.uri
        d["description"] = self.description
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NiktoVuln":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = NiktoVuln(id=d["id"], osvdb_id=d["osvdb_id"], method=d["method"], iplink=d["iplink"],
                        namelink=d["namelink"], uri=d["uri"], description=d["description"])
        return obj

    @staticmethod
    def from_json_file(json_file_path: str) -> "NiktoVuln":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NiktoVuln.from_dict(json.loads(json_str))



import csle_common.constants.constants as constants
import logging
from unittest.mock import MagicMock
from unittest.mock import patch
from csle_common.controllers.management_system_controller import (
    ManagementSystemController,
)
import subprocess


class TestManagementControllerSuite:
    """
    Test suite for management system controller
    """

    @patch("builtins.open")
    @patch("os.path.exists", return_value=True)
    def test_read_pid_file(self, mock_exists, mock_open) -> None:
        """
        Tests read_pid_file

        :param mock_exists: mock exists
        :param mock_open: mock open
        :return: None
        """
        path = "/path/to/pidfile"
        mock_file = mock_open.return_value
        mock_file.read.return_value = "1234"
        pid = ManagementSystemController.read_pid_file(path)
        assert pid == 1234
        mock_exists.assert_called_once_with(path)
        mock_open.assert_called_once_with(path, "r")

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.read_pid_file")
    def test_is_prometheus_running(self, mock_read_pid_file) -> None:
        """
        Tests the method that checks if prometheus is running on the host

        :param mock_read_pid_file: mock read_pid_file
        :return: None
        """
        mock_read_pid_file.return_value = -1
        assert not ManagementSystemController.is_prometheus_running()

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.read_pid_file")
    def test_is_node_exporter_running(self, mock_read_pid_file) -> None:
        """
        Tests the method that checks if node_exporter is running on the host

        :param mock_read_pid_file: mock read_pid_file
        :return: None
        """
        mock_read_pid_file.return_value = -1
        assert not ManagementSystemController.is_node_exporter_running()

    @patch("subprocess.run")
    def test_is_nginx_running(self, mock_run) -> None:
        """
        Tests the method that checks if Nginx is running on the host

        :param mock_run: mock run
        :return: None
        """
        constants.COMMANDS.NGINX_STATUS = "nginx status"
        mock_run.return_value = MagicMock(stdout="Active: active (running)")
        assert ManagementSystemController.is_nginx_running()
        mock_run.assert_called_once_with(["nginx", "status"], capture_output=True, text=True)

    @patch("subprocess.run")
    def test_is_postgresql_running(self, mock_run) -> None:
        """
        Tests the method that checks if postgresql is running on the host

        :param mock_run: mock run
        :return: None
        """
        constants.COMMANDS.POSTGRESQL_STATUS = "postgresql status"
        mock_run.return_value = MagicMock(stdout="Active: active (running)")
        assert ManagementSystemController.is_postgresql_running()
        mock_run.assert_called_once_with(["postgresql", "status"], capture_output=True, text=True)

    @patch("subprocess.run")
    def test_is_docker_engine_running(self, mock_run) -> None:
        """
        Tests the method that checks if Docker engine is running on the host

        :param mock_run: mock run
        :return: None
        """
        constants.COMMANDS.DOCKER_ENGINE_STATUS = "docker engine status"
        mock_run.return_value = MagicMock(stdout="Active: active (running)")
        assert ManagementSystemController.is_docker_engine_running()
        mock_run.assert_called_once_with(["docker", "engine", "status"], capture_output=True, text=True)

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.read_pid_file")
    def test_is_flask_running(self, mock_read_pid_file) -> None:
        """
        Tests the method that checks if flask web server is running on the host

        :param mock_run: mock run
        :return: None
        """
        mock_read_pid_file.return_value = -1
        assert not ManagementSystemController.is_flask_running()

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_node_exporter_running")
    @patch("subprocess.Popen")
    def test_start_node_exporter(self, mock_popen, mock_is_running) -> None:
        """
        Tests the method thats starts the node exporter

        :param mock_popen: mock popen
        :param mock_is_running: mock is_node_exporter_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = True
        result = ManagementSystemController.start_node_exporter(logger)
        assert not result
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("Node exporter is already running")
        mock_popen.assert_not_called()

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_flask_running")
    @patch("subprocess.Popen")
    def test_start_flask(self, mock_popen, mock_is_running) -> None:
        """
        Tests the method that starts Flask REST API Server

        :param mock_popen: mock popen
        :param mock_is_running: mock is_node_exporter_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = True
        result = ManagementSystemController.start_flask(logger)
        assert not result
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("Flask is already running")
        mock_popen.assert_not_called()

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_node_exporter_running")
    @patch("subprocess.Popen")
    def test_stop_node_exporter(self, mock_popen, mock_is_running) -> None:
        """
        Tests the method that stops the node exporter

        :param mock_popen: mock popen
        :param mock_is_running: mock is_node_exporter_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = False
        result = ManagementSystemController.stop_node_exporter(logger)
        assert not result
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("Node exporter is not running")
        mock_popen.assert_not_called()

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_flask_running")
    @patch("subprocess.Popen")
    def test_stop_flask(self, mock_popen, mock_is_running) -> None:
        """
        Tests the method that stops the flask REST API

        :param mock_popen: mock popen
        :param mock_is_running: mock is_flask_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = False
        result = ManagementSystemController.stop_flask(logger)
        assert not result
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("Flask is not running")
        mock_popen.assert_not_called()

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_prometheus_running")
    @patch("subprocess.Popen")
    def test_start_prometheus(self, mock_popen, mock_is_running) -> None:
        """
        Tests the method that starts Prometheus

        :param mock_popen: mock popen
        :param mock_is_running: mock is_prometheus_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = True
        result = ManagementSystemController.start_prometheus(logger)
        assert not result
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("Prometheus is already running")
        mock_popen.assert_not_called()

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_prometheus_running")
    @patch("subprocess.Popen")
    def test_stop_prometheus(self, mock_popen, mock_is_running) -> None:
        """
        Tests the method that stops Prometheus

        :param mock_popen: mock popen
        :param mock_is_running: mock is_prometheus_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = False
        result = ManagementSystemController.stop_prometheus(logger)
        assert not result
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("Prometheus is not running")
        mock_popen.assert_not_called()

    @patch("docker.from_env")
    def test_is_cadvisor_running(self, mock_from_env) -> None:
        """
        Tests if cadvisor is running

        :param mock_from_env: mock from_env
        :return: None
        """
        constants.CONTAINER_IMAGES = MagicMock()  # type: ignore
        constants.CONTAINER_IMAGES.CADVISOR = "name"
        mock_client = MagicMock()
        mock_from_env.return_value = mock_client
        mock_container = MagicMock()
        mock_container.name = "name"
        mock_client.containers.list.return_value = [mock_container]
        assert ManagementSystemController.is_cadvisor_running()
        mock_from_env.assert_called_once()
        mock_client.containers.list.assert_called_once()

    @patch("docker.from_env")
    def test_is_pgadmin_running(self, mock_from_env) -> None:
        """
        Tests if Prometheus is running

        :param mock_from_env: mock from_env
        :return: None
        """
        constants.CONTAINER_IMAGES = MagicMock()  # type: ignore
        constants.CONTAINER_IMAGES.PGADMIN = "name"
        mock_client = MagicMock()
        mock_from_env.return_value = mock_client
        mock_container = MagicMock()
        mock_container.name = "name"
        mock_client.containers.list.return_value = [mock_container]
        assert ManagementSystemController.is_pgadmin_running()
        mock_from_env.assert_called_once()
        mock_client.containers.list.assert_called_once()

    @patch("docker.from_env")
    def test_stop_cadvisor(self, mock_from_env) -> None:
        """
        Tests if cadvisor is stopped

        :param mock_from_env: mock from_env
        :return: None
        """
        constants.CONTAINER_IMAGES.CADVISOR = "name"
        mock_client = MagicMock()
        mock_from_env.return_value = mock_client
        mock_container = MagicMock()
        mock_container.name = "name"
        mock_client.containers.list.return_value = [mock_container]
        assert ManagementSystemController.stop_cadvisor()
        mock_from_env.assert_called_once()
        mock_client.containers.list.assert_called_once()
        mock_container.stop.assert_called_once()

    @patch("docker.from_env")
    def test_stop_pgadmin(self, mock_from_env) -> None:
        """
        Tests if pgadmin is stopped

        :param mock_from_env: mock from_env
        :return: None
        """
        constants.CONTAINER_IMAGES.PGADMIN = "name"
        mock_client = MagicMock()
        mock_from_env.return_value = mock_client
        mock_container = MagicMock()
        mock_container.name = "name"
        mock_client.containers.list.return_value = [mock_container]
        assert ManagementSystemController.stop_pgadmin()
        mock_from_env.assert_called_once()
        mock_client.containers.list.assert_called_once()
        mock_container.stop.assert_called_once()

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_cadvisor_running")
    def test_start_cadvisor(self, mock_is_running) -> None:
        """
        Tests if cadvisor starts

        :param mock_is_running: mock is_cadvisor_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = True
        result = ManagementSystemController.start_cadvisor(logger)
        assert not result
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("cAdvisor is already running")

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_postgresql_running")
    def test_start_postgresql(self, mock_is_running) -> None:
        """
        Tests if postgresql starts

        :param mock_is_running: mock is_postgresql_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = True
        result = ManagementSystemController.start_postgresql(logger)
        assert result == (False, None, None)
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("PostgreSQL is already running")

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_postgresql_running")
    def test_stop_postgresql(self, mock_is_running) -> None:
        """
        Tests if postgresql stopped

        :param mock_is_running: mock is_postgresql_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = False
        result = ManagementSystemController.stop_postgresql(logger)
        assert result == (False, None, None)
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("PostgreSQL is not running")

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_nginx_running")
    def test_start_nginx(self, mock_is_running) -> None:
        """
        Tests if Nginx starts

        :param mock_is_running: mock is_nginx_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = True
        result = ManagementSystemController.start_nginx(logger)
        assert result == (False, None, None)
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("Nginx is already running")

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_nginx_running")
    def test_stop_nginx(self, mock_is_running) -> None:
        """
        Tests if Nginx stops

        :param mock_is_running: mock is_nginx_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = False
        result = ManagementSystemController.stop_nginx(logger)
        assert result == (False, None, None)
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("Nginx is not running")

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_docker_engine_running")
    def test_start_docker_engine(self, mock_is_running) -> None:
        """
        Tests if Docker engine starts

        :param mock_is_running: mock is_docker_engine_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = True
        result = ManagementSystemController.start_docker_engine(logger)
        assert result == (False, None, None)
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("The Docker engine is already running")

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_docker_engine_running")
    def test_stop_docker_engine(self, mock_is_running) -> None:
        """
        Tests if Docker engine stops

        :param mock_is_running: mock is_docker_engine_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = False
        result = ManagementSystemController.stop_docker_engine(logger)
        assert result == (False, None, None)
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("The Docker engine is not running")

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_pgadmin_running")
    def test_start_pgadmin(self, mock_is_running) -> None:
        """
        Tests if pgadmin starts

        :param mock_is_running: mock is_pgadmin_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = True
        result = ManagementSystemController.start_pgadmin(logger)
        assert not result
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("pgAdmin is already running")

    @patch("docker.from_env")
    def test_stop_grafana(self, mock_from_env) -> None:
        """
        Tests if grafana stops

        :param mock_from_env: mock from_env
        :return: None
        """
        constants.CONTAINER_IMAGES.GRAFANA = "name"
        mock_client = MagicMock()
        mock_from_env.return_value = mock_client
        mock_container = MagicMock()
        mock_container.name = "name"
        mock_client.containers.list.return_value = [mock_container]
        assert ManagementSystemController.stop_grafana()
        mock_from_env.assert_called_once()
        mock_client.containers.list.assert_called_once()
        mock_container.stop.assert_called_once()

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_grafana_running")
    def test_start_grafana(self, mock_is_running) -> None:
        """
        Tests if grafana starts

        :param mock_is_running: mock is_grafana_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = True
        result = ManagementSystemController.start_grafana(logger)
        assert not result
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("Grafana is already running")

    @patch("docker.from_env")
    def test_is_grafana_running(self, mock_from_env) -> None:
        """
        Tests if grafana is running

        :param mock_from_env: mock from_env
        :return: None
        """
        constants.CONTAINER_IMAGES.GRAFANA = "name"
        mock_client = MagicMock()
        mock_from_env.return_value = mock_client
        mock_container = MagicMock()
        mock_container.name = "name"
        mock_client.containers.list.return_value = [mock_container]
        assert ManagementSystemController.is_grafana_running()
        mock_from_env.assert_called_once()
        mock_client.containers.list.assert_called_once()

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.read_pid_file")
    def test_is_statsmanager_running(self, mock_read_pid_file) -> None:
        """
        Tests the method that checks if statsmanager is running on the host

        :param mock_read_pid_file: mock read_pid_file
        :return: None
        """
        mock_read_pid_file.return_value = -1
        assert not ManagementSystemController.is_statsmanager_running()

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_statsmanager_running")
    def test_start_docker_statsmanager(self, mock_is_running) -> None:
        """
        Tests the method that checks if docker statsmanager is running on the host

        :param mock_is_running: mock is_statsmanager_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = True
        result = ManagementSystemController.start_docker_statsmanager(
            logger=logger,
            log_file="docker_stats_manager.log",
            log_dir="/var/log/csle",
            max_workers=10,
            port=50046,
        )
        assert not result
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("The docker statsmanager is already running")

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_statsmanager_running")
    def test_stop_docker_statsmanager(self, mock_is_running) -> None:
        """
        Tests the method that stops docker statsmanager

        :param mock_is_running: mock is_statsmanager_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = False
        result = ManagementSystemController.stop_docker_statsmanager(logger=logger)
        assert not result
        mock_is_running.assert_called_once()
        logger.info.assert_called_with("The statsmanager is not running")

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.read_pid_file")
    @patch("subprocess.Popen")
    def test_stop_cluster_manager(self, mock_popen, mock_read_pid_file) -> None:
        """
        Tests the method that stops the local cluster manager

        :param mock_read_pid_file: mock read_pid_file
        :return: None
        """
        mock_read_pid_file.return_value = 1234
        mock_popen.return_value.communicate.return_value = (None, None)
        result = ManagementSystemController.stop_cluster_manager()
        assert result
        mock_read_pid_file.assert_called_once_with(constants.COMMANDS.CLUSTER_MANAGER_PIDFILE)
        mock_popen.assert_called_once_with(constants.COMMANDS.KILL_PROCESS.format(1234), stdout=subprocess.DEVNULL,
                                           shell=True)
        mock_popen.return_value.communicate.assert_called_once()

    @patch("psutil.pid_exists")
    def test_is_pid_running(self, mock_pid_exists) -> None:
        """
        Tests the method that checks if the given pid is running on the host

        :param mock_pid_exists: mock pid_exists

        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_pid_exists.return_value = True
        pid = 1234
        result = ManagementSystemController.is_pid_running(pid, logger)
        assert result
        logger.info.assert_called_once_with(f"Checking if PID: {pid} is running")
        mock_pid_exists.assert_called_once_with(pid)

    @patch("csle_common.controllers.management_system_controller.ManagementSystemController.is_pid_running")
    def test_stop_pid(self, mock_is_running) -> None:
        """
        Tests the method that stops a process with a given pid

        :param mock_is_running: mock is_pid_running
        :return: None
        """
        logger = MagicMock(spec=logging.Logger)
        mock_is_running.return_value = False
        pid = 1234
        result = ManagementSystemController.stop_pid(pid, logger)
        assert not result
        logger.info.assert_not_called()



import gpytorch


class GPRegressionModelWithGaussianNoise(gpytorch.models.ExactGP):
    """
    DTO representing a GPRegression model with Gaussian Noise
    """

    def __init__(self, train_x, train_y, likelihood):
        """
        Initializes the model

        :param train_x: the x-data points
        :param train_y: the y-data points
        :param likelihood: The Gaussian likelihood that defines the observational distribution.
                           Since we're using exact inference, the likelihood must be Gaussian.
        """
        super(GPRegressionModelWithGaussianNoise, self).__init__(train_x, train_y, likelihood)
        self.mean_module = gpytorch.means.ConstantMean()
        self.covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel())

    def forward(self, x) -> gpytorch.distributions.MultivariateNormal:
        """
        Compute the prior latent distribution on a given input using the mean and covariance functions

        :param x: the input
        :return: the prior
        """
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)



from typing import List, Dict, Any
from csle_collector.client_manager.dao.constant_arrival_config import ConstantArrivalConfig
from csle_collector.client_manager.dao.workflows_config import WorkflowsConfig
from csle_collector.client_manager.dao.workflow_markov_chain import WorkflowMarkovChain
from csle_collector.client_manager.dao.workflow_service import WorkflowService
from csle_collector.client_manager.dao.client import Client
from csle_common.dao.emulation_config.container_network import ContainerNetwork
from csle_common.util.general_util import GeneralUtil
from csle_base.json_serializable import JSONSerializable


class ClientPopulationConfig(JSONSerializable):
    """
    A DTO object representing the configuration of the client population of an emulation
    """

    def __init__(self, ip: str, networks: List[ContainerNetwork],
                 client_manager_port: int, client_manager_log_file: str,
                 client_manager_log_dir: str, client_manager_max_workers: int,
                 clients: List[Client], workflows_config: WorkflowsConfig,
                 client_time_step_len_seconds: int = 1,
                 docker_gw_bridge_ip: str = "", physical_host_ip: str = ""):
        """
        Creates a ClientPopulationConfig DTO Object

        :param ip: the ip of the client container
        :param networks: a list of networks in the emulation that are accessible for external networks
        :param client_time_step_len_seconds: time-step length to measure the arrival process
        :param client_manager_log_file: the log file of the client manager
        :param client_manager_log_dir: the log dir of the client manager
        :param client_manager_max_workers: the maximum number of GRPC workers for the client manager
        :param docker_gw_bridge_ip: IP to reach the container from the host network
        :param physical_host_ip: IP of the physical host where the container is running
        :param clients: list of client configurations
        :param workflows_config: the workflows configurations
        """
        self.networks = networks
        self.ip = ip
        self.client_manager_port = client_manager_port
        self.client_time_step_len_seconds = client_time_step_len_seconds
        self.client_manager_log_dir = client_manager_log_dir
        self.client_manager_log_file = client_manager_log_file
        self.client_manager_max_workers = client_manager_max_workers
        self.docker_gw_bridge_ip = docker_gw_bridge_ip
        self.physical_host_ip = physical_host_ip
        self.clients = clients
        self.workflows_config = workflows_config

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ClientPopulationConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = ClientPopulationConfig(
            ip=d["ip"],
            networks=list(map(lambda x: ContainerNetwork.from_dict(x), d["networks"])),
            client_manager_port=d["client_manager_port"],
            client_time_step_len_seconds=d["client_time_step_len_seconds"],
            client_manager_log_dir=d["client_manager_log_dir"], client_manager_log_file=d["client_manager_log_file"],
            client_manager_max_workers=d["client_manager_max_workers"],
            docker_gw_bridge_ip=d["docker_gw_bridge_ip"], physical_host_ip=d["physical_host_ip"],
            clients=list(map(lambda x: Client.from_dict(x), d["clients"])),
            workflows_config=WorkflowsConfig.from_dict(d["workflows_config"])
        )
        return obj

    def no_clients(self) -> "ClientPopulationConfig":
        """
        :return: A version of the config with no clients
        """
        return ClientPopulationConfig(
            ip=self.ip,
            networks=self.networks,
            client_manager_port=self.client_manager_port,
            client_time_step_len_seconds=self.client_time_step_len_seconds,
            client_manager_log_file="client_manager.log", client_manager_log_dir="/", client_manager_max_workers=10,
            docker_gw_bridge_ip=self.docker_gw_bridge_ip, physical_host_ip=self.physical_host_ip,
            clients=[Client(id=0, arrival_config=ConstantArrivalConfig(lamb=0), mu=0, exponential_service_time=True,
                            workflow_distribution=[1])],
            workflows_config=WorkflowsConfig(
                workflow_markov_chains=[WorkflowMarkovChain(
                    initial_state=0, transition_matrix=[[0.8, 0.2], [0, 1]], id=0)],
                workflow_services=[WorkflowService(ips_and_commands=[('localhost', "echo ' '")], id=0)])
        )

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["ip"] = self.ip
        d["networks"] = list(map(lambda x: x.to_dict(), self.networks))
        d["client_manager_port"] = self.client_manager_port
        d["client_time_step_len_seconds"] = self.client_time_step_len_seconds
        d["client_manager_log_file"] = self.client_manager_log_file
        d["client_manager_log_dir"] = self.client_manager_log_dir
        d["client_manager_max_workers"] = self.client_manager_max_workers
        d["docker_gw_bridge_ip"] = self.docker_gw_bridge_ip
        d["physical_host_ip"] = self.physical_host_ip
        d["clients"] = list(map(lambda x: x.to_dict(), self.clients))
        d["workflows_config"] = self.workflows_config.to_dict()
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"ip:{self.ip}, " \
               f"networks:{list(map(lambda x: str(x), self.networks))}, " \
               f"client_manager_port: {self.client_manager_port}, " \
               f"client_time_step_len_seconds: {self.client_time_step_len_seconds}," \
               f"client_manager_log_file: {self.client_manager_log_file}, " \
               f"client_manager_log_dir: {self.client_manager_log_dir}, " \
               f"client_manager_max_workers: {self.client_manager_max_workers}, " \
               f"docker_gw_bridge_ip:{self.docker_gw_bridge_ip}, physical_host_ip: {self.physical_host_ip}," \
               f"clients:{list(map(lambda x: str(x), self.clients))}, workflows_config: {self.workflows_config}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "ClientPopulationConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ClientPopulationConfig.from_dict(json.loads(json_str))

    def copy(self) -> "ClientPopulationConfig":
        """
        :return: a copy of the DTO
        """
        return ClientPopulationConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "ClientPopulationConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.ip = GeneralUtil.replace_first_octet_of_ip(ip=config.ip, ip_first_octet=ip_first_octet)
        config.networks = list(map(lambda x: x.create_execution_config(ip_first_octet=ip_first_octet),
                                   config.networks))
        config.workflows_config = config.workflows_config.create_execution_config(ip_first_octet=ip_first_octet)
        return config



from typing import Dict, Any
from csle_common.util.general_util import GeneralUtil
from csle_base.json_serializable import JSONSerializable


class ContainerNetwork(JSONSerializable):
    """
    DTO representing an IP network of virtual containers
    """

    def __init__(self, name: str, subnet_mask: str, bitmask: str, subnet_prefix: str, interface: str = "eth0"):
        """
        Initializes the DTO

        :param name: the name of the network
        :param subnet_mask: the subnet mask of the network
        :param bitmask: the bitmask of the network
        :param subnet_prefix: the subnet prefix of the network
        :param interface: (optional) the interface of the network
        """
        self.name = name
        self.subnet_mask = subnet_mask
        self.bitmask = bitmask
        self.subnet_prefix = subnet_prefix
        self.interface = interface

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ContainerNetwork":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = ContainerNetwork(
            name=d["name"], subnet_mask=d["subnet_mask"], subnet_prefix=d["subnet_prefix"],
            interface=d["interface"], bitmask=d["bitmask"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d = {}
        d["name"] = self.name
        d["subnet_mask"] = self.subnet_mask
        d["subnet_prefix"] = self.subnet_prefix
        d["interface"] = self.interface
        d["bitmask"] = self.bitmask
        return d

    def __str__(self):
        """
        :return: a string representation of the object
        """
        return f"name:{self.name}, subnet_mask:{self.subnet_mask}, subnet_prefix: {self.subnet_prefix}, " \
               f"interface: {self.interface}, bitmask: {self.bitmask}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "ContainerNetwork":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ContainerNetwork.from_dict(json.loads(json_str))

    def copy(self) -> "ContainerNetwork":
        """
        :return: a copy of the DTO
        """
        return ContainerNetwork.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "ContainerNetwork":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.name = config.name + f"_{ip_first_octet}"
        config.subnet_mask = GeneralUtil.replace_first_octet_of_ip(ip=config.subnet_mask,
                                                                   ip_first_octet=ip_first_octet)
        config.subnet_prefix = GeneralUtil.replace_first_octet_of_ip(ip=config.subnet_mask,
                                                                     ip_first_octet=ip_first_octet)
        return config

    @staticmethod
    def schema() -> "ContainerNetwork":
        """
        :return: get the schema of the DTO
        """
        return ContainerNetwork(name="", subnet_mask="", bitmask="", subnet_prefix="", interface="")



from typing import List, Dict, Any
from csle_common.dao.simulation_config.env_parameter import EnvParameter
from csle_base.json_serializable import JSONSerializable


class EnvParametersConfig(JSONSerializable):
    """
    DTO representing the the configuration of custom environment parameters of a simulation
    """

    def __init__(self, parameters: List[EnvParameter]):
        """
        Initializes the DTO

        :param parameters: the list of parameters
        """
        self.parameters = parameters

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EnvParametersConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = EnvParametersConfig(parameters=list(map(lambda x: EnvParameter.from_dict(x), d["parameters"])))
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the DTO
        """
        d = {}
        d["parameters"] = list(map(lambda x: x.to_dict(), self.parameters))
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the DTO
        """
        return f"parameters: {list(map(lambda x: str(x), self.parameters))}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "EnvParametersConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EnvParametersConfig.from_dict(json.loads(json_str))



from enum import IntEnum


class PolicyType(IntEnum):
    """
    Enum representing the different policy types in CSLE
    """
    TABULAR = 0
    ALPHA_VECTORS = 1
    DQN = 2
    FNN_W_SOFTMAX = 3
    MIXED_MULTI_THRESHOLD = 4
    PPO = 5
    RANDOM = 6
    VECTOR = 7
    MULTI_THRESHOLD = 8
    LINEAR_THRESHOLD = 9
    MIXED_PPO_POLICY = 10
    LINEAR_TABULAR = 11
    MIXED_LINEAR_TABULAR = 12
    C51 = 13



from typing import Dict, Any
from csle_common.dao.simulation_config.simulation_env_input_config import SimulationEnvInputConfig
from csle_common.dao.simulation_config.players_config import PlayersConfig
from csle_common.dao.simulation_config.state_space_config import StateSpaceConfig
from csle_common.dao.simulation_config.joint_action_space_config import JointActionSpaceConfig
from csle_common.dao.simulation_config.joint_observation_space_config import JointObservationSpaceConfig
from csle_common.dao.simulation_config.time_step_type import TimeStepType
from csle_common.dao.simulation_config.reward_function_config import RewardFunctionConfig
from csle_common.dao.simulation_config.transition_operator_config import TransitionOperatorConfig
from csle_common.dao.simulation_config.observation_function_config import ObservationFunctionConfig
from csle_common.dao.simulation_config.initial_state_distribution_config import InitialStateDistributionConfig
from csle_common.dao.simulation_config.env_parameters_config import EnvParametersConfig
from csle_base.json_serializable import JSONSerializable


class SimulationEnvConfig(JSONSerializable):
    """
    A DTO class representing the configuration of a simulation environment
    """

    def __init__(self, name: str, descr: str, version: str,
                 gym_env_name: str,
                 simulation_env_input_config: SimulationEnvInputConfig,
                 players_config: PlayersConfig, state_space_config: StateSpaceConfig,
                 joint_action_space_config: JointActionSpaceConfig,
                 joint_observation_space_config: JointObservationSpaceConfig, time_step_type: TimeStepType,
                 reward_function_config: RewardFunctionConfig, transition_operator_config: TransitionOperatorConfig,
                 observation_function_config: ObservationFunctionConfig,
                 initial_state_distribution_config: InitialStateDistributionConfig,
                 env_parameters_config: EnvParametersConfig, plot_transition_probabilities: bool = False,
                 plot_observation_function: bool = False, plot_reward_function: bool = False) -> None:
        """
        Initializes the DTO

        :param name: the name of the simulation
        :param gym_env_name: the name of the OpenAI gym environment
        :param descr: the description of the simulation
        :param simulation_env_input_config: the input configuration to the simulation
        :param players_config: the players configuration of the simulation
        :param state_space_config: the state space configuration of the simulation
        :param joint_action_space_config: the joint action space configuration of the simulation
        :param joint_observation_space_config: the joint observation space config of the simulation
        :param time_step_type: the time step type of the simulation
        :param reward_function_config: the reward function configuration of the of the simulation
        :param transition_operator_config: the transition operator configuration of the simulation
        :param observation_function_config: the observation function configuration of the simulation
        :param emulation_statistic_id: the id of the emulation statistic
        :param initial_state_distribution_config: the initial state distribution configuration of the simulation
        :param version: the version of the environment
        :param env_parameters_config: parameters that are not part of the state but that the poliy depends on.
        :param plot_transition_probabilities: boolean parameter whether to plot transition probabilities or not
        :param plot_observation_function: boolean parameter whether to plot the observation function or not
        :param plot_reward_function: boolean parameter whether to plot the reward function or not
        """
        self.name = name
        self.descr = descr
        self.simulation_env_input_config = simulation_env_input_config
        self.players_config = players_config
        self.state_space_config = state_space_config
        self.joint_action_space_config = joint_action_space_config
        self.joint_observation_space_config = joint_observation_space_config
        self.time_step_type = time_step_type
        self.reward_function_config = reward_function_config
        self.transition_operator_config = transition_operator_config
        self.observation_function_config = observation_function_config
        self.initial_state_distribution_config = initial_state_distribution_config
        self.version = version
        self.gym_env_name = gym_env_name
        self.image = None
        self.id = -1
        self.env_parameters_config = env_parameters_config
        self.plot_transition_probabilities = plot_transition_probabilities
        self.plot_observation_function = plot_observation_function
        self.plot_reward_function = plot_reward_function

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "SimulationEnvConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        from gym_csle_stopping_game.dao.stopping_game_config import StoppingGameConfig
        from gym_csle_stopping_game.dao.stopping_game_defender_pomdp_config import StoppingGameDefenderPomdpConfig
        from gym_csle_stopping_game.dao.stopping_game_attacker_mdp_config import StoppingGameAttackerMdpConfig
        from gym_csle_apt_game.dao.apt_game_config import AptGameConfig
        from gym_csle_apt_game.dao.apt_game_defender_pomdp_config import AptGameDefenderPomdpConfig
        from gym_csle_apt_game.dao.apt_game_attacker_mdp_config import AptGameAttackerMdpConfig
        from gym_csle_intrusion_response_game.dao.intrusion_response_game_local_pomdp_defender_config \
            import IntrusionResponseGameLocalPOMDPDefenderConfig
        from gym_csle_intrusion_response_game.dao.intrusion_response_game_local_pomdp_attacker_config \
            import IntrusionResponseGameLocalPOMDPAttackerConfig
        from gym_csle_intrusion_response_game.dao.workflow_intrusion_response_pomdp_defender_config \
            import WorkflowIntrusionResponsePOMDPDefenderConfig
        from gym_csle_intrusion_response_game.dao.workflow_intrusion_response_pomdp_attacker_config \
            import WorkflowIntrusionResponsePOMDPAttackerConfig
        from csle_tolerance.dao.intrusion_recovery_pomdp_config import IntrusionRecoveryPomdpConfig
        from csle_tolerance.dao.intrusion_response_cmdp_config import IntrusionResponseCmdpConfig
        from gym_csle_cyborg.dao.csle_cyborg_config import CSLECyborgConfig
        input_config = None
        parse_functions = [StoppingGameConfig.from_dict, StoppingGameAttackerMdpConfig.from_dict,
                           StoppingGameDefenderPomdpConfig.from_dict,
                           AptGameConfig.from_dict, AptGameAttackerMdpConfig.from_dict,
                           AptGameDefenderPomdpConfig.from_dict,
                           IntrusionResponseGameLocalPOMDPDefenderConfig.from_dict,
                           IntrusionResponseGameLocalPOMDPAttackerConfig.from_dict,
                           WorkflowIntrusionResponsePOMDPDefenderConfig.from_dict,
                           WorkflowIntrusionResponsePOMDPAttackerConfig.from_dict,
                           IntrusionRecoveryPomdpConfig.from_dict,
                           IntrusionResponseCmdpConfig.from_dict, CSLECyborgConfig.from_dict]
        for parse_fun in parse_functions:
            try:
                input_config = parse_fun(d["simulation_env_input_config"])
                break
            except Exception:
                pass
        if input_config is None:
            raise ValueError(f"Could not parse the input configuration: \n{d['simulation_env_input_config']}")
        obj = SimulationEnvConfig(
            name=d["name"], descr=d["descr"],
            simulation_env_input_config=input_config,
            players_config=PlayersConfig.from_dict(d["players_config"]),
            state_space_config=StateSpaceConfig.from_dict(d["state_space_config"]),
            joint_action_space_config=JointActionSpaceConfig.from_dict(d["joint_action_space_config"]),
            joint_observation_space_config=JointObservationSpaceConfig.from_dict(d["joint_observation_space_config"]),
            time_step_type=TimeStepType(d["time_step_type"]),
            reward_function_config=RewardFunctionConfig.from_dict(d["reward_function_config"]),
            transition_operator_config=TransitionOperatorConfig.from_dict(d["transition_operator_config"]),
            observation_function_config=ObservationFunctionConfig.from_dict(d["observation_function_config"]),
            initial_state_distribution_config=InitialStateDistributionConfig.from_dict(
                d["initial_state_distribution_config"]),
            version=d["version"], gym_env_name=d["gym_env_name"],
            env_parameters_config=EnvParametersConfig.from_dict(d["env_parameters_config"]),
            plot_transition_probabilities=d["plot_transition_probabilities"],
            plot_observation_function=d["plot_observation_function"],
            plot_reward_function=d["plot_reward_function"]
        )
        obj.id = d["id"]
        obj.image = d["image"]
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["name"] = self.name
        d["descr"] = self.descr
        d["simulation_env_input_config"] = self.simulation_env_input_config.to_dict()
        d["players_config"] = self.players_config.to_dict()
        d["state_space_config"] = self.state_space_config.to_dict()
        d["joint_action_space_config"] = self.joint_action_space_config.to_dict()
        d["joint_observation_space_config"] = self.joint_observation_space_config.to_dict()
        d["time_step_type"] = self.time_step_type.value
        d["reward_function_config"] = self.reward_function_config.to_dict()
        d["transition_operator_config"] = self.transition_operator_config.to_dict()
        d["observation_function_config"] = self.observation_function_config.to_dict()
        d["initial_state_distribution_config"] = self.initial_state_distribution_config.to_dict()
        d["version"] = self.version
        d["gym_env_name"] = self.gym_env_name
        d["id"] = self.id
        d["image"] = self.image
        d["env_parameters_config"] = self.env_parameters_config.to_dict()
        d["plot_transition_probabilities"] = self.plot_transition_probabilities
        d["plot_observation_function"] = self.plot_observation_function
        d["plot_reward_function"] = self.plot_reward_function
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"name: {self.name}, descr: {self.descr}, " \
               f"simulation_env_input_config: {self.simulation_env_input_config}, " \
               f"players_config: {self.players_config}, state_space_config: {self.state_space_config}," \
               f"joint_action_space_config: {self.joint_action_space_config}, " \
               f"joint_observation_space_config: {self.joint_observation_space_config}, " \
               f"time_step_type: {self.time_step_type}, reward_function_config: {self.reward_function_config}, " \
               f"transition_operator_config: {self.transition_operator_config}, " \
               f"observation_function_config: {self.observation_function_config}, " \
               f"initial_state_distribution_config: {self.initial_state_distribution_config}," \
               f"version: {self.version}, gym_env_name: {self.gym_env_name}, id: {self.id}," \
               f"env_parameters_config: {self.env_parameters_config}," \
               f"plot_observation_function: {self.plot_observation_function}," \
               f"plot_transition_probabilities: {self.plot_transition_probabilities}," \
               f"plot_reward_function: {self.plot_reward_function}"

    def copy(self) -> "SimulationEnvConfig":
        """
        :return: a copy of the DTO
        """
        return self.from_dict(self.to_dict())

    @staticmethod
    def from_json_file(json_file_path: str) -> "SimulationEnvConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return SimulationEnvConfig.from_dict(json.loads(json_str))



from typing import Optional, Dict, Any
from csle_common.dao.emulation_config.credential import Credential
from csle_base.json_serializable import JSONSerializable


class EmulationConnectionObservationState(JSONSerializable):
    """
    A DTO representing a connection observation in the emulation
    """

    def __init__(self, conn, credential: Credential, root: bool, service: str, port: int, tunnel_thread=None,
                 tunnel_port: Optional[int] = None, interactive_shell=None,
                 proxy: Optional["EmulationConnectionObservationState"] = None, ip: Optional[str] = None):
        """
        Intializes the DTO

        :param conn: the connection object
        :param credential: the credential of the connection
        :param root: whether the connection is root or not
        :param service: the service of the connection
        :param port: the port of the connection
        :param tunnel_thread: the tunnel thread for the connection
        :param tunnel_port: the tunnel port of the connection
        :param interactive_shell: an interactive shell of the connection
        :param proxy: a proxy for the connection
        :param ip: the ip of the connection
        """
        self.conn = conn
        self.credential = credential
        self.root = root
        self.port = port
        self.service = service
        self.tunnel_thread = tunnel_thread
        self.tunnel_port = tunnel_port
        self.interactive_shell = interactive_shell
        self.proxy = proxy
        self.ip = ip

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationConnectionObservationState":
        """
        Converts a dict representation of the object to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = EmulationConnectionObservationState(
            conn=None, credential=Credential.from_dict(d["credential"]), root=d["root"], port=d["port"],
            service=d["service"], tunnel_port=d["tunnel_port"], tunnel_thread=None, interactive_shell=None, ip=d["ip"],
            proxy=None)
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict represnetation of the object
        """
        d: Dict[str, Any] = {}
        d["credential"] = self.credential.to_dict()
        d["root"] = self.root
        d["port"] = self.port
        d["service"] = self.service
        d["ip"] = self.ip
        d["tunnel_port"] = self.tunnel_port
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the connection observation
        """
        return "credential:{},root:{},service:{},port:{}".format(self.credential, self.root, self.service, self.port)

    def __eq__(self, other) -> bool:
        """
        Checks for equality with another connection

        :param other: the other connection to compare with
        :return: True if equal, otherwise False
        """
        if not isinstance(other, EmulationConnectionObservationState):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return (self.credential.username == other.credential.username and self.root == other.root
                and self.service == other.service and self.port == other.port and self.ip == other.ip)

    def __hash__(self) -> int:
        """
        :return: a hash representation of the object
        """
        return (hash(self.credential.username) + 31 * hash(self.root) + 31 * hash(self.service) + 31 *
                hash(self.port) + 31 * hash(self.ip))

    def cleanup(self) -> None:
        """
        Utility function for cleaning up the connection.

        :return: None
        """

        if self.tunnel_thread is not None:
            try:
                self.tunnel_thread.shutdown()
            except Exception:
                pass
            self.tunnel_thread = None
        if self.interactive_shell is not None:
            try:
                self.interactive_shell.close()
            except Exception:
                pass
            self.interactive_shell = None
        if self.conn is not None:
            try:
                self.conn.close()
            except Exception:
                pass
            self.conn = None
        if self.proxy is not None:
            try:
                self.proxy.cleanup()
            except Exception:
                pass
            self.proxy = None

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationConnectionObservationState":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationConnectionObservationState.from_dict(json.loads(json_str))

    def num_attributes(self) -> int:
        """
        :return: The number of attribute of the DTO
        """
        return 9 + self.credential.num_attributes()

    @staticmethod
    def schema() -> "EmulationConnectionObservationState":
        """
        :return: get the schema of the DTO
        """
        return EmulationConnectionObservationState(conn=None, credential=Credential.schema(), root=False, service="",
                                                   port=-1, tunnel_thread=None, tunnel_port=-1,
                                                   interactive_shell=False, proxy=None, ip="")



from typing import Dict, Any
from abc import abstractmethod
from csle_base.json_serializable import JSONSerializable


class SimulationEnvInputConfig(JSONSerializable):
    """
    Abstract class representing the input to a simulation
    """

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        pass

    @staticmethod
    @abstractmethod
    def from_dict(d: Dict[str, Any]) -> "SimulationEnvInputConfig":
        """
        Converts a dict representation of the object to an instance

        :param d: the dict to convert
        :return: the converted instance
        """
        pass

    @staticmethod
    @abstractmethod
    def from_json_file(json_file_path: str) -> "SimulationEnvInputConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        pass



from csle_common.dao.datasets.statistics_dataset import StatisticsDataset
from csle_common.dao.datasets.traces_dataset import TracesDataset


class TestDatasetsDaoSuite:
    """
    Test suite for datasets data access objects (DAOs)
    """

    def test_traces_dataset(self, example_traces_dataset: TracesDataset) -> None:
        """
        Tests creation and dict conversion of the TracesDataset DAO

        :param test_traces_dataset: an example TracesDataset
        :return: None
        """
        assert isinstance(example_traces_dataset.to_dict(), dict)
        assert isinstance(TracesDataset.from_dict(example_traces_dataset.to_dict()), TracesDataset)
        assert TracesDataset.from_dict(example_traces_dataset.to_dict()).to_dict() == example_traces_dataset.to_dict()
        assert TracesDataset.from_dict(example_traces_dataset.to_dict()) == example_traces_dataset

    def test_statistics_dataset(self, example_statistics_dataset: StatisticsDataset) -> None:
        """
        Tests creation and dict conversion of the StatisticsDataset DAO

        :param example_statistics_dataset: an example StatisticsDataset
        :return: None
        """
        assert isinstance(example_statistics_dataset.to_dict(), dict)
        assert isinstance(StatisticsDataset.from_dict(example_statistics_dataset.to_dict()), StatisticsDataset)
        assert StatisticsDataset.from_dict(example_statistics_dataset.to_dict()).to_dict() == \
               example_statistics_dataset.to_dict()
        assert StatisticsDataset.from_dict(example_statistics_dataset.to_dict()) == example_statistics_dataset



from typing import List, Dict, Any
from csle_common.dao.emulation_config.node_users_config import NodeUsersConfig
from csle_base.json_serializable import JSONSerializable


class UsersConfig(JSONSerializable):
    """
    A DTO object representing the users configuration of an emulation environment
    """

    def __init__(self, users_configs: List[NodeUsersConfig]):
        """
        Initializes the DTO

        :param users_configs: the list of node users configuration
        """
        self.users_configs = users_configs

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "UsersConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = UsersConfig(
            users_configs=list(map(lambda x: NodeUsersConfig.from_dict(x), d["users_configs"]))
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d = {}
        d["users_configs"] = list(map(lambda x: x.to_dict(), self.users_configs))
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return "users:{}".format(",".join(list(map(lambda x: str(x), self.users_configs))))

    def get_root_usernames(self, ips: List[str]) -> List[str]:
        """
        Gets the root usernames for a list of ips

        :param ips: the list of ips to get the root usernames for
        :return: the list of root usernames
        """
        root_usernames: List[str] = []
        for users_config in self.users_configs:
            if users_config.ip in ips:
                for user in users_config.users:
                    if user.root:
                        root_usernames.append(user.username)
        return root_usernames

    @staticmethod
    def from_json_file(json_file_path: str) -> "UsersConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return UsersConfig.from_dict(json.loads(json_str))

    def copy(self) -> "UsersConfig":
        """
        :return: a copy of the DTO
        """
        return UsersConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "UsersConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.users_configs = list(map(lambda x: x.create_execution_config(ip_first_octet=ip_first_octet),
                                        config.users_configs))
        return config



from typing import Dict, Any, Union
from csle_base.json_serializable import JSONSerializable
from csle_collector.five_g_core_manager.five_g_core_manager_pb2 import FiveGSubscriberDTO


class FiveGSubscriberConfig(JSONSerializable):
    """
    DTO representing the configuration of a 5G subscriber
    """

    def __init__(self, imsi: str, key: str, opc: str, amf: str, sqn: int) -> None:
        """
        Initializes the DTO

        :param imsi: the imsi of the subscriber (International Mobile Subscriber Identity)
        :param key: the private key of the subscriber
        :param opc: the operator key
        :param amf: the AMF (Authentication Management Field)
        :param sqn: the sequence number (SQN)
        """
        self.imsi = imsi
        self.key = key
        self.opc = opc
        self.amf = amf
        self.sqn = sqn

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "FiveGSubscriberConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = FiveGSubscriberConfig(imsi=d["imsi"], key=d["key"], opc=d["opc"], amf=d["amf"], sqn=d["sqn"])
        return obj

    def to_dict(self) -> Dict[str, Union[str, int]]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Union[str, int]] = {}
        d["imsi"] = self.imsi
        d["key"] = self.key
        d["opc"] = self.opc
        d["amf"] = self.amf
        d["sqn"] = self.sqn
        return d

    def to_subscriber_dto(self) -> FiveGSubscriberDTO:
        """
        Converts the object to a SubscriberDTO

        :return: the created SubscriberDTO
        """
        return FiveGSubscriberDTO(imsi=self.imsi, key=self.key, opc=self.opc, amf=self.amf, sqn=self.sqn)

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"imsi: {self.imsi}, key: {self.key}, opc: {self.opc}, amf: {self.amf}, sqn: {self.sqn}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "FiveGSubscriberConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return FiveGSubscriberConfig.from_dict(json.loads(json_str))

    def copy(self) -> "FiveGSubscriberConfig":
        """
        :return: a copy of the DTO
        """
        return FiveGSubscriberConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "FiveGSubscriberConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        return config

    @staticmethod
    def schema() -> "FiveGSubscriberConfig":
        """
        :return: get the schema of the DTO
        """
        return FiveGSubscriberConfig(
            imsi="001010123456780", key="00112233445566778899aabbccddeeff", opc="63BFA50EE6523365FF14C1F45F88737D",
            amf="8000", sqn=10
        )



from typing import Dict, Any
from csle_base.json_serializable import JSONSerializable


class AgentLog(JSONSerializable):
    """
    DTO Representing the Agent Log
    """

    def __init__(self):
        """
        Initializes the log
        """
        self.log = []

    def add_entry(self, msg) -> None:
        """
        Adds an entry to the log

        :param msg: the msg to add to the log
        :return: None
        """
        self.log.append(msg)

    def reset(self) -> None:
        """
        Resets the log

        :return: None
        """
        self.log = []

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "AgentLog":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = AgentLog()
        obj.log = d["log"]
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d = {}
        d["log"] = self.log
        return d

    @staticmethod
    def from_json_file(json_file_path: str) -> "AgentLog":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return AgentLog.from_dict(json.loads(json_str))



from typing import List, Dict, Any
from csle_common.util.general_util import GeneralUtil
from csle_base.json_serializable import JSONSerializable


class NodeTrafficConfig(JSONSerializable):
    """
    A DTO object representing the traffic configuration of an individual container in an emulation
    """

    def __init__(self, ip: str, commands: List[str], traffic_manager_log_file: str, traffic_manager_log_dir: str,
                 traffic_manager_max_workers: int, traffic_manager_port: int = 50043,
                 docker_gw_bridge_ip: str = "", physical_host_ip: str = ""):
        """
        Creates a NodeTrafficConfig DTO Object

        :param ip: the ip of the node that generate the traffic
        :param commands: the commands used to generate the traffic
        :param traffic_manager_log_file: the file name to write logs of the traffic manager
        :param traffic_manager_log_dir: the directory to save the log file of the traffic manager
        :param traffic_manager_max_workers: the maximum number of gRPC workers for the traffic manager
        :param traffic_manager_port: the port of the traffic manager
        :param docker_gw_bridge_ip: IP to reach the container from the host network
        :param physical_host_ip: IP of the physical host where the container is running
        """
        self.ip = ip
        self.commands = commands
        self.traffic_manager_port = traffic_manager_port
        self.traffic_manager_log_file = traffic_manager_log_file
        self.traffic_manager_log_dir = traffic_manager_log_dir
        self.traffic_manager_max_workers = traffic_manager_max_workers
        self.docker_gw_bridge_ip = docker_gw_bridge_ip
        self.physical_host_ip = physical_host_ip

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NodeTrafficConfig":
        """
        Converts a dict representation into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = NodeTrafficConfig(
            ip=d["ip"], commands=d["commands"], traffic_manager_port=d["traffic_manager_port"],
            traffic_manager_max_workers=d["traffic_manager_max_workers"],
            traffic_manager_log_file=d["traffic_manager_log_file"],
            traffic_manager_log_dir=d["traffic_manager_log_dir"],
            docker_gw_bridge_ip=d["docker_gw_bridge_ip"],
            physical_host_ip=d["physical_host_ip"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["ip"] = self.ip
        d["commands"] = self.commands
        d["traffic_manager_port"] = self.traffic_manager_port
        d["traffic_manager_max_workers"] = self.traffic_manager_max_workers
        d["traffic_manager_log_file"] = self.traffic_manager_log_file
        d["traffic_manager_log_dir"] = self.traffic_manager_log_dir
        d["docker_gw_bridge_ip"] = self.docker_gw_bridge_ip
        d["physical_host_ip"] = self.physical_host_ip
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"ip:{self.ip}, commands:{self.commands}, traffic_manager_port: {self.traffic_manager_port}, " \
               f"traffic_manager_log_file: {self.traffic_manager_log_file}, " \
               f"traffic_manager_max_workers: {self.traffic_manager_max_workers}, " \
               f"traffic_manager_log_file_dir: {self.traffic_manager_log_dir}," \
               f" docker_gw_bridge_ip:{self.docker_gw_bridge_ip}," \
               f" physical_host_ip: {self.physical_host_ip}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "NodeTrafficConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NodeTrafficConfig.from_dict(json.loads(json_str))

    def copy(self) -> "NodeTrafficConfig":
        """
        :return: a copy of the DTO
        """
        return NodeTrafficConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "NodeTrafficConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.ip = GeneralUtil.replace_first_octet_of_ip(ip=config.ip, ip_first_octet=ip_first_octet)
        return config



from typing import Dict, Any
from csle_common.dao.emulation_config.transport_protocol import TransportProtocol
from csle_common.dao.emulation_config.network_service import NetworkService
from csle_base.json_serializable import JSONSerializable


class EmulationPortObservationState(JSONSerializable):
    """
    DTO Representation a port observation in the emulation
    """

    def __init__(self, port: int, open: bool, service: str, protocol: TransportProtocol, http_enum: str = "",
                 http_grep: str = "", vulscan: str = "", version: str = "", fingerprint: str = ""):
        """
        Initializes the DTO

        :param port: the port
        :param open: whether the port is open or not (boolean)
        :param service: the service of the port
        :param protocol: the protocol of the port
        :param http_enum: the HTTP enum result on the port
        :param http_grep: the HTTP grep result on the port
        :param vulscan: the vulscan result of the port
        :param version: the version of the port
        :param fingerprint: the fingerprint output of the port
        """
        self.port = port
        self.open = open
        self.service = service
        self.protocol = protocol
        self.http_enum = http_enum
        self.http_grep = http_grep
        self.vulscan = vulscan
        self.version = version
        self.fingerprint = fingerprint

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["port"] = self.port
        d["open"] = self.open
        d["service"] = self.service
        d["protocol"] = self.protocol
        d["http_enum"] = self.http_enum
        d["http_grep"] = self.http_grep
        d["vulscan"] = self.vulscan
        d["version"] = self.version
        d["fingerprint"] = self.fingerprint
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationPortObservationState":
        """
        Converts a dict representation of the object to an instance

        :param d: the dict to convert
        :return: the created object instance
        """
        obj = EmulationPortObservationState(
            port=d["port"],
            open=d["open"],
            service=d["service"],
            protocol=d["protocol"],
            http_enum=d["http_enum"],
            http_grep=d["http_grep"],
            vulscan=d["vulscan"],
            version=d["version"],
            fingerprint=d["fingerprint"]
        )
        return obj

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"port:{self.port}, open:{self.open}, service:{self.service}, protocol:{self.protocol}, " \
               f"http_enum:{self.http_enum}, http_grep:{self.http_grep}, vulscan:{self.vulscan}, " \
               f"version:{self.version}, fingerprint:{self.fingerprint}"

    def to_network_service(self) -> NetworkService:
        """
        Converts the object into a network service representation

        :return: the network service representation
        """
        service = NetworkService(protocol=self.protocol, port=self.port, name=self.service, credentials=[])
        return service

    @staticmethod
    def from_network_service(network_service: NetworkService,
                             service_lookup: Dict[str, str]) -> "EmulationPortObservationState":
        """
        Converts a network service into a port observation state

        :param network_service: the network service to convert
        :param service_lookup: a lookup table for converting between service names and service ids
        :return:
        """
        if network_service.name in service_lookup:
            s = service_lookup[network_service.name]
        else:
            s = service_lookup["unknown"]
        port = EmulationPortObservationState(port=network_service.port, open=True,
                                             service=s,
                                             protocol=network_service.protocol)
        return port

    def copy(self) -> "EmulationPortObservationState":
        """
        :return: a copy of the object
        """
        c = EmulationPortObservationState(
            port=self.port, open=self.open, service=self.service, protocol=self.protocol, http_enum=self.http_enum,
            http_grep=self.http_grep, vulscan=self.vulscan, version=self.version, fingerprint=self.fingerprint
        )
        return c

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationPortObservationState":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationPortObservationState.from_dict(json.loads(json_str))

    def num_attributes(self) -> int:
        """
        :return: The number of attribute of the DTO
        """
        return 9

    @staticmethod
    def schema() -> "EmulationPortObservationState":
        """
        :return: get the schema of the DTO
        """
        return EmulationPortObservationState(port=-1, open=True, service="", protocol=TransportProtocol.TCP,
                                             http_enum="", http_grep="", vulscan="", version="", fingerprint="")



from typing import List, Dict, Any
from sklearn.mixture import GaussianMixture
from scipy.stats import norm
import numpy as np
from csle_base.json_serializable import JSONSerializable


class GaussianMixtureConditional(JSONSerializable):
    """
    A DTO representing a Gaussian Mixture Conditional Distribution
    """

    def __init__(self, conditional_name: str, metric_name: str, num_mixture_components: int,
                 dim: int, mixtures_means: List[List[float]],
                 mixtures_covariance_matrix: List[List[List[float]]], mixture_weights: List[float],
                 sample_space: List[int]) -> None:
        """
        Initializes the DTO

        :param conditional_name: the name of the conditional
        :param num_mixture_components: the number of mixture components
        :param dim: the dimension of the distribution, i.e., if it is multivariate
        :param mixtures_means: the means of the mixtures
        :param mixtures_covariance_matrix: the covariance matrices of the mixtures
        :param mixture_weights: the mixture weights
        :param sample_space: the sampĺe space
        """
        self.conditional_name = conditional_name
        self.dim = dim
        self.num_mixture_components = num_mixture_components
        self.mixtures_means = mixtures_means
        self.mixtures_covariance_matrix = mixtures_covariance_matrix
        self.mixture_weights = mixture_weights
        self.metric_name = metric_name
        self.sample_space = sample_space
        self.weighted_mixture_distributions: List[List[float]] = []
        self.generate_distributions()
        self.combined_distribution: List[float] = []

    def generate_distributions(self) -> None:
        """
        Generates the combined mixture distribution by taking a convex combination of the Gaussians.

        :return: None
        """
        self.sample_space.sort()
        dists = []
        for weight, mean, covar in zip(self.mixture_weights, self.mixtures_means, self.mixtures_covariance_matrix):
            dists.append(list(weight * norm.pdf(self.sample_space, mean, np.sqrt(covar)).ravel()))
        self.weighted_mixture_distributions = dists
        combined_dist = np.zeros(len(self.sample_space))
        for dist in dists:
            d_arr = np.array(dist)
            combined_dist = combined_dist + d_arr
        self.combined_distribution = list(combined_dist)

    def generate_distributions_for_samples(self, samples: List[float], normalize: bool = False):
        """
        Generates a distribution over a given list of samples

        :param samples: the samples to generate the distribution for
        :param normalize: a boolean flag indicating whether the density distribution should be normalized or not
        :return: the mixture distribution over the given samples.
        """
        samples.sort()
        dists = []
        for weight, mean, covar in zip(self.mixture_weights, self.mixtures_means, self.mixtures_covariance_matrix):
            density_dist = list(weight * norm.pdf(samples, mean, np.sqrt(covar)).ravel())
            dists.append(density_dist)
        combined_density_dist = np.zeros(len(samples))
        for density_dist in dists:
            d_arr = np.array(density_dist)
            combined_density_dist = combined_density_dist + d_arr
        combined_density_dist_1 = list(combined_density_dist)
        if normalize:
            combined_prob_dist = list(np.array(combined_density_dist_1) * (1 / sum(combined_density_dist_1)))
        else:
            combined_prob_dist = combined_density_dist_1
        return combined_prob_dist

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "GaussianMixtureConditional":
        """
        Converts a dict representation of the DTO into an instance

        :param d: the dict to convert
        :return: the converted instance
        """
        return GaussianMixtureConditional(
            conditional_name=d["conditional_name"],
            num_mixture_components=d["num_mixture_components"],
            dim=d["dim"], mixtures_means=d["mixture_means"],
            mixtures_covariance_matrix=d["mixtures_covariance_matrix"],
            mixture_weights=d["mixture_weights"], metric_name=d["metric_name"],
            sample_space=d["sample_space"]
        )

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the DTO
        """
        d: Dict[str, Any] = {}
        d["conditional_name"] = self.conditional_name
        d["dim"] = self.dim
        d["num_mixture_components"] = self.num_mixture_components
        d["mixture_means"] = self.mixtures_means
        d["mixtures_covariance_matrix"] = self.mixtures_covariance_matrix
        d["mixture_weights"] = self.mixture_weights
        d["metric_name"] = self.metric_name
        d["sample_space"] = self.sample_space
        d["weighted_mixture_distributions"] = self.weighted_mixture_distributions
        d["combined_distribution"] = self.combined_distribution
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the DTO
        """
        return f"conditional_name:{self.conditional_name}, num_mixture_components: {self.num_mixture_components}, " \
               f"dim: {self.dim}, mixtures_means: {self.mixtures_means}, " \
               f"mixtures_covariance_matrix: {self.mixtures_covariance_matrix}, " \
               f"mixture_weights: {self.mixture_weights}," \
               f"metric_name: {self.metric_name}, sample space: {self.sample_space}, " \
               f"combined distribution: {self.combined_distribution}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "GaussianMixtureConditional":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return GaussianMixtureConditional.from_dict(json.loads(json_str))

    @staticmethod
    def from_sklearn_gaussian_mixture(gmm: GaussianMixture, conditional_name: str, metric_name: str,
                                      num_components: int, sample_space: List[int],
                                      dim: int = 1) -> "GaussianMixtureConditional":
        """
        Creates the DTO from a Gaussian mixture fitted with sklearn

        :param gmm: the sklearn model
        :param conditional_name: the name of the conditional
        :param metric_name: the metric name
        :param num_components: the number of components of the mixture
        :param dim: the dimension of the mixture
        :param sample_space: the sample space
        :return: a GaussianMixtureConditional instance
        """
        mixture_weights = list(gmm.weights_)
        means = list(gmm.means_.tolist())
        covariances = list(gmm.covariances_.tolist())
        return GaussianMixtureConditional(
            conditional_name=conditional_name, metric_name=metric_name, num_mixture_components=num_components,
            mixtures_means=means, mixtures_covariance_matrix=covariances, mixture_weights=mixture_weights, dim=dim,
            sample_space=sample_space)



import subprocess
from unittest.mock import patch, MagicMock
from csle_common.controllers.resource_constraints_controller import ResourceConstraintsController


class TestResourceConstraintsControllerSuite:
    """
    Test suite for ResourceConstraintsController
    """

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.disconnect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("subprocess.Popen")
    def test_apply_resource_constraints(
            self, mock_popen, mock_execute_ssh_cmd, mock_disconnect_admin, mock_connect_admin) -> None:
        """
        Test the method that creates users in an emulation environment according to a specified users-configuration


        :param mock_popen: mock_popen
        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :param mock_disconnect_admin: mock_disconnect_admin
        :param mock_connect_admin: mock_connect_admin
        :return: None
        """
        emulation_env_config = MagicMock()
        network_config = MagicMock()
        network_config.interface = "eth0"
        network_config.packet_delay_ms = 10
        network_config.packet_delay_jitter_ms = 5
        network_config.packet_delay_distribution = "normal"
        network_config.loss_gemodel_p = 0.1
        network_config.loss_gemodel_r = 0.1
        network_config.loss_gemodel_h = 0.9
        network_config.loss_gemodel_k = 0.9
        network_config.packet_duplicate_percentage = 0.1
        network_config.packet_duplicate_correlation_percentage = 0.1
        network_config.packet_corrupt_percentage = 0.1
        network_config.packet_reorder_percentage = 0.1
        network_config.packet_reorder_correlation_percentage = 0.1
        network_config.packet_reorder_gap = 5
        network_config.rate_limit_mbit = 10
        network_config.limit_packets_queue = 100

        node_resource_config = MagicMock()
        node_resource_config.physical_host_ip = "192.168.1.1"
        node_resource_config.docker_gw_bridge_ip = "192.168.1.2"
        node_resource_config.container_name = "container_1"
        node_resource_config.available_memory_gb = 2
        node_resource_config.num_cpus = 1
        node_resource_config.ips_and_network_configs = [("192.168.1.2", network_config)]

        emulation_env_config.resources_config.node_resources_configurations = [node_resource_config]
        emulation_env_config.kafka_config.resources = node_resource_config
        emulation_env_config.sdn_controller_config.resources = node_resource_config
        emulation_env_config.connections = {"192.168.1.2": MagicMock()}
        physical_server_ip = "192.168.1.1"
        mock_execute_ssh_cmd.return_value = ("output", "error", 0)
        logger = MagicMock()
        ResourceConstraintsController.apply_resource_constraints(
            emulation_env_config=emulation_env_config, physical_server_ip=physical_server_ip, logger=logger)
        mock_connect_admin.assert_any_call(emulation_env_config=emulation_env_config, ip="192.168.1.2")
        expected_docker_update_cmd = "docker update --memory=2G --cpus=1 container_1"
        mock_popen.assert_any_call(expected_docker_update_cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL,
                                   shell=True)
        mock_execute_ssh_cmd.assert_called()
        mock_disconnect_admin.assert_any_call(emulation_env_config=emulation_env_config)



import json
import numpy as np


class NpEncoder(json.JSONEncoder):
    """
    Encoder for Numpy arrays to JSON
    """
    def default(self, obj):
        """
        The default encoding of numpy datatypes for JSON

        :param obj: the object to encoded
        :return: the encoded object
        """
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NpEncoder, self).default(obj)



import logging
from unittest.mock import MagicMock, patch
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.controllers.elk_controller import ELKController
import csle_common.constants.constants as constants
import csle_collector.elk_manager.elk_manager_pb2_grpc
import csle_collector.elk_manager.elk_manager_pb2
import csle_collector.elk_manager.query_elk_manager
import csle_collector.elk_manager.elk_manager_util


class TestElkControllerSuite:
    """
    Test suite for elk controller
    """

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("time.sleep", return_value=None)
    def test_start_stop_elk_manager(self, mock_sleep, mock_execute_ssh_cmd, mock_connect_admin) -> None:
        """
        Test case for starting or stopping the ELK manager

        :param mock_disconnect_admin: Mocked disconnect_admin method
        :param mock_execute_ssh_cmd: Mocked execute_ssh_cmd method
        :param mock_connect_admin: Mocked connect_admin
        :param mock_sleep: Mocked sleep method
        :return: None
        """
        mock_connect_admin.return_value = None
        mock_execute_ssh_cmd.return_value = ("output", "error", 0)
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        elk_config = MagicMock()
        emulation_env_config.elk_config = elk_config
        emulation_env_config.elk_config.container.docker_gw_bridge_ip = "172.17.0.1"
        emulation_env_config.elk_config.get_connection.return_value = MagicMock()
        logger = MagicMock(spec=logging.Logger)
        ELKController.start_elk_manager(emulation_env_config, logger)
        mock_connect_admin.assert_called_once_with(emulation_env_config=emulation_env_config, ip="172.17.0.1",
                                                   create_producer=False)
        mock_execute_ssh_cmd.assert_called()
        ELKController.stop_elk_manager(emulation_env_config, logger)
        assert 2 == mock_connect_admin.call_count
        assert 3 == mock_execute_ssh_cmd.call_count

    @patch("csle_common.controllers.elk_controller.ELKController.start_elk_manager")
    @patch("csle_common.controllers.elk_controller.ELKController.get_elk_status_by_port_and_ip")
    def test_get_elk_status(self, mock_get_elk_status_by_port_and_ip, mock_start_elk_manager) -> None:
        """
        Test case for querying the ELKManager about the status of the ELK stack

        :param mock_get_elk_status_by_port_and_ip: mocked get_elk_status_by_port_and_ip method
        :param mock_start_elk_manager: mocked start_elk_manager method
        """
        mock_start_elk_manager.return_value = None
        expected_elk_dto = MagicMock(spec=csle_collector.elk_manager.elk_manager_pb2.ElkDTO)
        mock_get_elk_status_by_port_and_ip.return_value = expected_elk_dto
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        elk_config = MagicMock()
        emulation_env_config.elk_config = elk_config
        emulation_env_config.elk_config.container.docker_gw_bridge_ip = "172.17.0.1"
        emulation_env_config.elk_config.elk_manager_port = 5601
        logger = MagicMock(spec=logging.Logger)
        result = ELKController.get_elk_status(emulation_env_config, logger)
        mock_start_elk_manager.assert_called_once_with(emulation_env_config=emulation_env_config, logger=logger)
        mock_get_elk_status_by_port_and_ip.assert_called_once_with(ip="172.17.0.1", port=5601)
        assert result == expected_elk_dto

    @patch("csle_collector.elk_manager.query_elk_manager.get_elk_status")
    @patch("csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub")
    @patch("grpc.insecure_channel")
    def test_get_elk_status_by_port_and_ip(self, mock_insecure_channel, mock_ElkManagerStub, mock_get_elk_status) \
            -> None:
        """
        Test case for querying the ElkManager about the status of the ELK stack

        :param mock_insecure_channel: mocked insecure_channel method
        :param mock_ElkManagerStub: mocked ElkManagerStub method
        :param mock_get_elk_status: mocked get_elk_status method
        """
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_ElkManagerStub.return_value = mock_stub
        expected_elk_dto = MagicMock(spec=csle_collector.elk_manager.elk_manager_pb2.ElkDTO)
        mock_get_elk_status.return_value = expected_elk_dto
        ip = "172.17.0.1"
        port = 5601
        timeout = 5
        result = ELKController.get_elk_status_by_port_and_ip(ip, port, timeout=timeout)
        mock_insecure_channel.assert_called_once_with(f"{ip}:{port}", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_ElkManagerStub.assert_called_once_with(mock_channel)
        mock_get_elk_status.assert_called_once_with(mock_stub, timeout=timeout)
        assert result == expected_elk_dto

    @patch("csle_collector.elk_manager.query_elk_manager.stop_elk")
    @patch("csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub")
    @patch("grpc.insecure_channel")
    @patch("csle_common.controllers.elk_controller.ELKController.start_elk_manager")
    def test_stop_elk_stack(self, mock_start_elk_manager, mock_insecure_channel, mock_ElkManagerStub, mock_stop_elk) \
            -> None:
        """
        Test case for requesting the ELKManager to stop the ELK server

        :param mock_start_elk_manager: mocked start_elk_manager method
        :param mock_insecure_channel: mocked insecure_channel method
        :param mock_ElkManagerStub: mocked ElkManagerStub
        :param mock_stop_elk: mocked stop_elk method
        """
        mock_start_elk_manager.return_value = None
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_ElkManagerStub.return_value = mock_stub
        expected_elk_dto = MagicMock(spec=csle_collector.elk_manager.elk_manager_pb2.ElkDTO)
        mock_stop_elk.return_value = expected_elk_dto
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        elk_config = MagicMock()
        emulation_env_config.elk_config = elk_config
        emulation_env_config.elk_config.container.docker_gw_bridge_ip = "172.17.0.1"
        emulation_env_config.elk_config.elk_manager_port = 5601
        logger = MagicMock(spec=logging.Logger)
        result = ELKController.stop_elk_stack(emulation_env_config, logger)
        mock_insecure_channel.assert_called_once_with("172.17.0.1:5601", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_ElkManagerStub.assert_called_once_with(mock_channel)
        mock_stop_elk.assert_called_once_with(mock_stub)
        assert result == expected_elk_dto

    @patch("csle_collector.elk_manager.query_elk_manager.start_elk")
    @patch("csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub")
    @patch("grpc.insecure_channel")
    @patch("csle_common.controllers.elk_controller.ELKController.start_elk_manager")
    def test_start_elk_stack(self, mock_start_elk_manager, mock_insecure_channel, mock_ElkManagerStub,
                             mock_start_elk) -> None:
        """
        Test case for requesting the ELKManager to start the ELK server

        :param mock_start_elk_manager: mocked start_elk_manager method
        :param mock_insecure_channel: mocked insecure_channel method
        :param mock_ElkManagerStub: mocked ElkManagerStub
        :param mock_start_elk: mocked start_elk method
        """
        mock_start_elk_manager.return_value = None
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_ElkManagerStub.return_value = mock_stub
        expected_elk_dto = MagicMock(spec=csle_collector.elk_manager.elk_manager_pb2.ElkDTO)
        mock_start_elk.return_value = expected_elk_dto
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        elk_config = MagicMock()
        emulation_env_config.elk_config = elk_config
        emulation_env_config.elk_config.container.docker_gw_bridge_ip = "172.17.0.1"
        emulation_env_config.elk_config.container.physical_host_ip = "192.168.0.1"
        emulation_env_config.elk_config.elk_manager_port = 5601
        logger = MagicMock(spec=logging.Logger)
        physical_server_ip = "192.168.0.1"
        result = ELKController.start_elk_stack(emulation_env_config, physical_server_ip, logger)
        mock_insecure_channel.assert_called_once_with("172.17.0.1:5601", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_ElkManagerStub.assert_called_once_with(mock_channel)
        mock_start_elk.assert_called_once_with(mock_stub)
        assert result == expected_elk_dto

    @patch("csle_collector.elk_manager.query_elk_manager.start_elastic")
    @patch("csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub")
    @patch("grpc.insecure_channel")
    @patch("csle_common.controllers.elk_controller.ELKController.start_elk_manager")
    def test_start_elastic(self, mock_start_elk_manager, mock_insecure_channel, mock_ElkManagerStub,
                           mock_start_elastic) -> None:
        """
        Test case for requesting the ELKManager to start elasticsearch

        :param mock_start_elk_manager: mocked start_elk_manager method
        :param mock_insecure_channel: mocked insecure_channel method
        :param mock_ElkManagerStub: mocked ElkManagerStub
        :param mock_start_elastic: mocked start_elastic method
        """
        mock_start_elk_manager.return_value = None
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_ElkManagerStub.return_value = mock_stub
        expected_elk_dto = MagicMock(spec=csle_collector.elk_manager.elk_manager_pb2.ElkDTO)
        mock_start_elastic.return_value = expected_elk_dto
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        elk_config = MagicMock()
        emulation_env_config.elk_config = elk_config
        emulation_env_config.elk_config.container.docker_gw_bridge_ip = "172.17.0.1"
        emulation_env_config.elk_config.elk_manager_port = 5601
        logger = MagicMock(spec=logging.Logger)
        result = ELKController.start_elastic(emulation_env_config, logger)
        mock_insecure_channel.assert_called_once_with("172.17.0.1:5601", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_ElkManagerStub.assert_called_once_with(mock_channel)
        mock_start_elastic.assert_called_once_with(mock_stub)
        assert result == expected_elk_dto

    @patch("csle_collector.elk_manager.query_elk_manager.start_kibana")
    @patch("csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub")
    @patch("grpc.insecure_channel")
    @patch("csle_common.controllers.elk_controller.ELKController.start_elk_manager")
    def test_start_kibana(self, mock_start_elk_manager, mock_insecure_channel, mock_ElkManagerStub,
                          mock_start_kibana) -> None:
        """
        Test case for requesting the ELKManager to start kibana

        :param mock_start_elk_manager: mocked start_elk_manager method
        :param mock_insecure_channel: mocked insecure_channel method
        :param mock_ElkManagerStub: mocked ElkManagerStub
        :param mock_start_kibana: mocked start_kibana method
        """
        mock_start_elk_manager.return_value = None
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_ElkManagerStub.return_value = mock_stub
        expected_elk_dto = MagicMock(spec=csle_collector.elk_manager.elk_manager_pb2.ElkDTO)
        mock_start_kibana.return_value = expected_elk_dto
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        elk_config = MagicMock()
        emulation_env_config.elk_config = elk_config
        emulation_env_config.elk_config.container.docker_gw_bridge_ip = "172.17.0.1"
        emulation_env_config.elk_config.elk_manager_port = 5601
        logger = MagicMock(spec=logging.Logger)
        result = ELKController.start_kibana(emulation_env_config, logger)
        mock_insecure_channel.assert_called_once_with("172.17.0.1:5601", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_ElkManagerStub.assert_called_once_with(mock_channel)
        mock_start_kibana.assert_called_once_with(mock_stub)
        assert result == expected_elk_dto

    @patch("csle_collector.elk_manager.query_elk_manager.start_logstash")
    @patch("csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub")
    @patch("grpc.insecure_channel")
    @patch("csle_common.controllers.elk_controller.ELKController.start_elk_manager")
    def test_start_logstash(self, mock_start_elk_manager, mock_insecure_channel, mock_ElkManagerStub,
                            mock_start_logstash) -> None:
        """
        Test case for requesting the ELKManager to start logstash

        :param mock_start_elk_manager: mocked start_elk_manager method
        :param mock_insecure_channel: mocked insecure_channel method
        :param mock_ElkManagerStub: mocked ElkManagerStub
        :param mock_start_logstash: mocked start_logstash method
        """
        mock_start_elk_manager.return_value = None
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_ElkManagerStub.return_value = mock_stub
        expected_elk_dto = MagicMock(spec=csle_collector.elk_manager.elk_manager_pb2.ElkDTO)
        mock_start_logstash.return_value = expected_elk_dto
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        elk_config = MagicMock()
        emulation_env_config.elk_config = elk_config
        emulation_env_config.elk_config.container.docker_gw_bridge_ip = "172.17.0.1"
        emulation_env_config.elk_config.elk_manager_port = 5601
        logger = MagicMock(spec=logging.Logger)
        result = ELKController.start_logstash(emulation_env_config, logger)
        mock_insecure_channel.assert_called_once_with("172.17.0.1:5601", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_ElkManagerStub.assert_called_once_with(mock_channel)
        mock_start_logstash.assert_called_once_with(mock_stub)
        assert result == expected_elk_dto

    @patch("csle_collector.elk_manager.query_elk_manager.stop_elastic")
    @patch("csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub")
    @patch("grpc.insecure_channel")
    @patch("csle_common.controllers.elk_controller.ELKController.start_elk_manager")
    def test_stop_elastic(self, mock_start_elk_manager, mock_insecure_channel, mock_ElkManagerStub,
                          mock_stop_elastic) -> None:
        """
        Test case for requesting the ELKManager to stop elasticsearch

        :param mock_start_elk_manager: mocked start_elk_manager method
        :param mock_insecure_channel: mocked insecure_channel method
        :param mock_ElkManagerStub: mocked ElkManagerStub
        :param mock_stop_elastic: mocked stop_elastic method
        """
        mock_start_elk_manager.return_value = None
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_ElkManagerStub.return_value = mock_stub
        expected_elk_dto = MagicMock(spec=csle_collector.elk_manager.elk_manager_pb2.ElkDTO)
        mock_stop_elastic.return_value = expected_elk_dto
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        elk_config = MagicMock()
        emulation_env_config.elk_config = elk_config
        emulation_env_config.elk_config.container.docker_gw_bridge_ip = "172.17.0.1"
        emulation_env_config.elk_config.elk_manager_port = 5601
        logger = MagicMock(spec=logging.Logger)
        result = ELKController.stop_elastic(emulation_env_config, logger)
        mock_insecure_channel.assert_called_once_with("172.17.0.1:5601", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_ElkManagerStub.assert_called_once_with(mock_channel)
        mock_stop_elastic.assert_called_once_with(mock_stub)
        assert result == expected_elk_dto

    @patch("csle_collector.elk_manager.query_elk_manager.stop_kibana")
    @patch("csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub")
    @patch("grpc.insecure_channel")
    @patch("csle_common.controllers.elk_controller.ELKController.start_elk_manager")
    def test_stop_kibana(self, mock_start_elk_manager, mock_insecure_channel, mock_ElkManagerStub,
                         mock_stop_kibana) -> None:
        """
        Test case for requesting the ELKManager to stop kibana

        :param mock_start_elk_manager: mocked start_elk_manager method
        :param mock_insecure_channel: mocked insecure_channel method
        :param mock_ElkManagerStub: mocked ElkManagerStub
        :param mock_stop_kibana: mocked stop_kibana method
        """
        mock_start_elk_manager.return_value = None
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_ElkManagerStub.return_value = mock_stub
        expected_elk_dto = MagicMock(spec=csle_collector.elk_manager.elk_manager_pb2.ElkDTO)
        mock_stop_kibana.return_value = expected_elk_dto
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        elk_config = MagicMock()
        emulation_env_config.elk_config = elk_config
        emulation_env_config.elk_config.container.docker_gw_bridge_ip = "172.17.0.1"
        emulation_env_config.elk_config.elk_manager_port = 5601
        logger = MagicMock(spec=logging.Logger)
        result = ELKController.stop_kibana(emulation_env_config, logger)
        mock_insecure_channel.assert_called_once_with("172.17.0.1:5601", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_ElkManagerStub.assert_called_once_with(mock_channel)
        mock_stop_kibana.assert_called_once_with(mock_stub)
        assert result == expected_elk_dto

    @patch("csle_collector.elk_manager.query_elk_manager.stop_logstash")
    @patch("csle_collector.elk_manager.elk_manager_pb2_grpc.ElkManagerStub")
    @patch("grpc.insecure_channel")
    @patch("csle_common.controllers.elk_controller.ELKController.start_elk_manager")
    def test_stop_logstash(self, mock_start_elk_manager, mock_insecure_channel, mock_ElkManagerStub,
                           mock_stop_logstash) -> None:
        """
        Test case for requesting the ELKManager to stop logstash

        :param mock_start_elk_manager: mocked start_elk_manager method
        :param mock_insecure_channel: mocked insecure_channel method
        :param mock_ElkManagerStub: mocked ElkManagerStub
        :param mock_stop_logstash: mocked stop_logstash method
        """
        mock_start_elk_manager.return_value = None
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_ElkManagerStub.return_value = mock_stub
        expected_elk_dto = MagicMock(spec=csle_collector.elk_manager.elk_manager_pb2.ElkDTO)
        mock_stop_logstash.return_value = expected_elk_dto
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        elk_config = MagicMock()
        emulation_env_config.elk_config = elk_config
        emulation_env_config.elk_config.container.docker_gw_bridge_ip = "172.17.0.1"
        emulation_env_config.elk_config.elk_manager_port = 5601
        logger = MagicMock(spec=logging.Logger)
        result = ELKController.stop_logstash(emulation_env_config, logger)
        mock_insecure_channel.assert_called_once_with("172.17.0.1:5601", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_ElkManagerStub.assert_called_once_with(mock_channel)
        mock_stop_logstash.assert_called_once_with(mock_stub)
        assert result == expected_elk_dto

    def test_get_elk_managers_ips(self) -> None:
        """
        Test case for extracting the IPS of the ELK managers in a given emulation
        """
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        elk_config = MagicMock()
        emulation_env_config.elk_config = elk_config
        emulation_env_config.elk_config.container.docker_gw_bridge_ip = "172.17.0.1"
        result = ELKController.get_elk_managers_ips(emulation_env_config)
        assert result == ["172.17.0.1"]

    def test_get_elk_managers_ports(self) -> None:
        """
        Test case for extracting the ports of the ELK managers in a given emulation
        """
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        elk_config = MagicMock()
        emulation_env_config.elk_config = elk_config
        emulation_env_config.elk_config.elk_manager_port = 5601
        result = ELKController.get_elk_managers_ports(emulation_env_config)
        assert result == [5601]

    @patch("csle_common.controllers.elk_controller.ELKController.get_elk_managers_ips")
    @patch("csle_common.controllers.elk_controller.ELKController.get_elk_managers_ports")
    @patch("csle_common.controllers.elk_controller.ELKController.get_elk_status_by_port_and_ip")
    @patch("csle_collector.elk_manager.elk_manager_util.ElkManagerUtil.elk_dto_empty")
    def test_get_elk_managers_infos(self, mock_elk_dto_empty, mock_get_elk_status_by_port_and_ip,
                                    mock_get_elk_managers_ports, mock_get_elk_managers_ips) -> None:
        """
        Test case for extracting the infomation of the ELK managers for a given emulation

        :param mock_elk_dto_empty: mocked elk_dto_empty method
        :param mock_get_elk_status_by_port_and_ip: mocked get_elk_status_by_port_and_ip method
        :param mock_get_elk_managers_ports: mocked get_elk_managers_ports
        :param mock_get_elk_managers_ips: mocked get_elk_managers_ips
        """
        mock_get_elk_managers_ips.return_value = ["172.17.0.1"]
        mock_get_elk_managers_ports.return_value = [5601]
        mock_elk_status = MagicMock()
        mock_get_elk_status_by_port_and_ip.return_value = mock_elk_status
        mock_elk_dto_empty.return_value = MagicMock()
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        elk_config = MagicMock()
        emulation_env_config.elk_config = elk_config
        emulation_env_config.elk_config.container.docker_gw_bridge_ip = "172.17.0.1"
        emulation_env_config.elk_config.elk_manager_port = 5601
        emulation_env_config.execution_id = 1
        emulation_env_config.name = "test_emulation"
        logger = MagicMock(spec=logging.Logger)
        active_ips = ["172.17.0.1"]
        physical_host_ip = "192.168.0.1"
        result = ELKController.get_elk_managers_info(emulation_env_config, active_ips, logger, physical_host_ip)
        mock_get_elk_managers_ips.assert_called_once_with(emulation_env_config=emulation_env_config)
        mock_get_elk_managers_ports.assert_called_once_with(emulation_env_config=emulation_env_config)
        assert result.ips == ["172.17.0.1"]
        assert result.ports == [5601]
        assert result.execution_id == 1
        assert result.emulation_name == "test_emulation"



from typing import List, Dict, Any, Union
import gpytorch
from numpy.typing import NDArray
import torch
from csle_common.dao.system_identification.gp_regression_model_with_gauissan_noise import \
    GPRegressionModelWithGaussianNoise
from csle_base.json_serializable import JSONSerializable


class GPConditional(JSONSerializable):
    """
    A DTO representing a Gaussian process conditional distribution
    """

    def __init__(self, conditional_name: str, metric_name: str,
                 sample_space: List[int],
                 observed_x: List[Union[float, int]], observed_y: List[Union[float, int]],
                 scale_parameter: float, noise_parameter: float) -> None:
        """
        Initializes the DTO

        :param conditional_name: the name of the conditional
        :param metric_name: the name of the metric
        :param sample_space: the sample space (the domain of the distribution)
        :param observed_x: the observed x samples
        :param observed_y: the observed y samples
        :param scale_parameter: the scale parameter for the scale kernel
        :param noise_parameter: the noise parameter for the Gaussian noise likelihood
        """
        self.conditional_name = conditional_name
        self.metric_name = metric_name
        self.sample_space = sample_space
        self.observed_x = observed_x
        self.observed_y = observed_y
        self.scale_parameter = scale_parameter
        self.noise_parameter = noise_parameter
        self.distribution: List[NDArray[Any]] = []

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "GPConditional":
        """
        Converts a dict representation of the DTO into an instance

        :param d: the dict to convert
        :return: the converted instance
        """
        return GPConditional(
            conditional_name=d["conditional_name"], metric_name=d["metric_name"],
            sample_space=d["sample_space"], observed_x=d["observed_x"], observed_y=d["observed_y"],
            scale_parameter=d["scale_parameter"], noise_parameter=d["noise_parameter"]
        )

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the DTO
        """
        d: Dict[str, Any] = {}
        d["conditional_name"] = self.conditional_name
        d["metric_name"] = self.metric_name
        d["sample_space"] = self.sample_space
        d["observed_x"] = self.observed_x
        d["observed_y"] = self.observed_y
        d["scale_parameter"] = self.scale_parameter
        d["noise_parameter"] = self.noise_parameter
        d["distribution"] = self.distribution
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the DTO
        """
        return f"conditional_name:{self.conditional_name}, metric_name: {self.metric_name}, " \
               f"sample_space: {self.sample_space}, observed_x: {self.observed_x}, observed_y: {self.observed_y}," \
               f"scale_parameter: {self.scale_parameter}, noise_parameter: {self.noise_parameter}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "GPConditional":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return GPConditional.from_dict(json.loads(json_str))

    def generate_distribution(self) -> None:
        """
        Updates the distribution

        :return: None
        """
        self.sample_space.sort()
        self.distribution = list(self.generate_distributions_for_samples(samples=self.sample_space).tolist())

    def generate_distributions_for_samples(self, samples) -> Any:
        """
        Generates distributions for a given sample

        :param samples: the sample
        :return: the created distributions
        """
        samples = torch.tensor(samples)
        likelihood = gpytorch.likelihoods.GaussianLikelihood()
        model = GPRegressionModelWithGaussianNoise(torch.tensor(self.observed_x), torch.tensor(self.observed_y),
                                                   likelihood)
        model.covar_module.base_kernel.lengthscale = torch.tensor(self.scale_parameter)
        model.likelihood.noise = torch.tensor(self.noise_parameter)

        model.eval()
        likelihood.eval()
        # Make predictions by feeding model through likelihood
        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            test_x = torch.tensor(samples)
            dist = likelihood(model(test_x)).mean.numpy()
            return dist



from typing import List, Dict, Any, Union
from csle_common.dao.emulation_config.node_traffic_config import NodeTrafficConfig
from csle_common.dao.emulation_config.client_population_config import ClientPopulationConfig
from csle_base.json_serializable import JSONSerializable


class TrafficConfig(JSONSerializable):
    """
    A DTO object representing the traffic configuration of an emulation environment
    """
    def __init__(self, node_traffic_configs: List[NodeTrafficConfig],
                 client_population_config: ClientPopulationConfig) -> None:
        """
        Initializes the DTO

        :param node_traffic_configs: the list of node traffic configurations
        :param client_population_config: the configuration of the client population
        """
        self.node_traffic_configs = node_traffic_configs
        self.client_population_config = client_population_config

    def get_node_traffic_config_by_ip(self, ip: str) -> Union[NodeTrafficConfig, None]:
        """
        Gets a node traffic config with a specific IP

        :param ip: the ip
        :return: the node traffic config or None
        """
        for node_traffic_config in self.node_traffic_configs:
            if node_traffic_config.ip == ip or ip == node_traffic_config.docker_gw_bridge_ip:
                return node_traffic_config
        return None

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "TrafficConfig":
        """
        Converts a dict representation of the object into a an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = TrafficConfig(
            node_traffic_configs=list(map(lambda x: NodeTrafficConfig.from_dict(x), d["node_traffic_configs"])),
            client_population_config=ClientPopulationConfig.from_dict(d["client_population_config"]))
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["node_traffic_configs"] = list(map(lambda x: x.to_dict(), self.node_traffic_configs))
        d["client_population_config"] = self.client_population_config.to_dict()
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"node_traffic_configs:{','.join(list(map(lambda x: str(x), self.node_traffic_configs)))}, " \
               f"client_population_config: {self.client_population_config}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "TrafficConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return TrafficConfig.from_dict(json.loads(json_str))

    def copy(self) -> "TrafficConfig":
        """
        :return: a copy of the DTO
        """
        return TrafficConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "TrafficConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.client_population_config = config.client_population_config.create_execution_config(
            ip_first_octet=ip_first_octet)
        config.node_traffic_configs = list(map(lambda x: x.create_execution_config(ip_first_octet=ip_first_octet),
                                               config.node_traffic_configs))
        return config



from typing import List, Dict, Tuple, Union, Optional, Any
import math
import random
import numpy as np
import csle_common.constants.constants as constants
from csle_common.dao.training.policy import Policy
from csle_common.dao.training.agent_type import AgentType
from csle_common.dao.simulation_config.state import State
from csle_common.dao.simulation_config.state_type import StateType
from csle_common.dao.training.player_type import PlayerType
from csle_common.dao.simulation_config.action import Action
from csle_common.dao.training.experiment_config import ExperimentConfig
from csle_common.dao.training.policy_type import PolicyType


class MultiThresholdStoppingPolicy(Policy):
    """
    A multi-threshold stopping policy
    """

    def __init__(self, theta: List[float], simulation_name: str, L: int, states: List[State], player_type: PlayerType,
                 actions: List[Action], experiment_config: Optional[ExperimentConfig], avg_R: float,
                 agent_type: AgentType, opponent_strategy: Optional["MultiThresholdStoppingPolicy"] = None):
        """
        Initializes the policy

        :param theta: the threshold vector
        :param simulation_name: the simulation name
        :param attacker: whether it is an attacker or not
        :param L: the number of stop actions
        :param states: list of states (required for computing stage policies)
        :param actions: list of actions
        :param experiment_config: the experiment configuration used for training the policy
        :param avg_R: the average reward of the policy when evaluated in the simulation
        :param agent_type: the agent type
        :param opponent_strategy: optionally an opponent strategy
        """
        super(MultiThresholdStoppingPolicy, self).__init__(agent_type=agent_type, player_type=player_type)
        self.theta = theta
        self.id = -1
        self.simulation_name = simulation_name
        self.L = L
        self.states = states
        self.actions = actions
        self.experiment_config = experiment_config
        self.avg_R = avg_R
        self.opponent_strategy = opponent_strategy
        self.policy_type = PolicyType.MULTI_THRESHOLD

    def action(self, o: List[float], deterministic: bool = True) -> int:
        """
        Multi-threshold stopping policy

        :param o: the current observation
        :param deterministic: boolean flag indicating whether the action selection should be deterministic
        :return: the selected action
        """
        if not self.player_type == PlayerType.ATTACKER:
            a, _ = self._defender_action(o=o)
            return a
        else:
            a, _ = self._attacker_action(o=o)
            return a

    def probability(self, o: List[float], a: int) -> float:
        """
        Probability of a given action

        :param o: the current observation
        :param a: a given action
        :return: the probability of a
        """
        if not self.player_type == PlayerType.ATTACKER:
            _, prob = self._defender_action(o=o)
            return prob
        else:
            _, prob = self._attacker_action(o=o)
            return prob

    def _attacker_action(self, o, defender_stopping_prob: Optional[float] = None) -> Tuple[int, float]:
        """
        Multi-threshold stopping policy of the attacker

        :param o: the input observation
        :param defender_stopping_prob: the defender's stop probability (optional)
        :return: the selected action (int) and its probability
        """
        s = o[2]
        l = int(o[0])
        if s == 2:
            return 0, 1.0
        if int(s * self.L + l - 1) <= len(self.theta):
            theta_val = self.theta[int(s * self.L + l - 1)]
        else:
            theta_val = self.theta[0]
        if defender_stopping_prob is None:
            if self.opponent_strategy is None:
                raise ValueError("The opponent strategy is None")
            a1, prob = self.opponent_strategy._defender_action(o=o)
            if a1 == 1:
                defender_stopping_prob = prob
            else:
                defender_stopping_prob = 1 - prob

        threshold = MultiThresholdStoppingPolicy.sigmoid(theta_val)
        if s == 0:
            a, attacker_action_prob = MultiThresholdStoppingPolicy.smooth_threshold_action_selection(
                threshold=threshold, b1=defender_stopping_prob, threshold_action=0, alternative_action=1, k=-20)
        elif s == 1:
            a, attacker_action_prob = MultiThresholdStoppingPolicy.smooth_threshold_action_selection(
                threshold=threshold, b1=defender_stopping_prob, threshold_action=1, alternative_action=0, k=-20)
        elif s == 2:
            a = 0
            attacker_action_prob = 1.0
        else:
            raise ValueError(f"Invalid state: {s}, valid states are: 0, 1, and 2")
        return a, attacker_action_prob

    def stage_policy(self, o: Union[List[int], List[float]]) -> List[List[float]]:
        """
        Gets the stage policy, i.e a |S|x|A| policy

        :param o: the latest observation
        :return: the |S|x|A| stage policy
        """
        b1 = o[1]
        l = int(o[0])
        threshold = MultiThresholdStoppingPolicy.sigmoid(self.theta[l - 1])
        if not self.player_type == PlayerType.ATTACKER:
            stage_policy = []
            for _ in self.states:
                stopping_probability = MultiThresholdStoppingPolicy.stopping_probability(b1=b1,
                                                                                         threshold=threshold, k=-20)
                stage_policy.append([1 - stopping_probability, stopping_probability])
            return stage_policy
        else:
            stage_policy = []
            if self.opponent_strategy is None:
                raise ValueError("The opponent strategy is None")
            a1, defender_stopping_probability = self.opponent_strategy._defender_action(o=o)
            if a1 == 0:
                defender_stopping_probability = 1 - defender_stopping_probability
            for s in self.states:
                if s.state_type != StateType.TERMINAL:
                    theta_val = self.theta[s.id * self.L + l - 1]
                    threshold = MultiThresholdStoppingPolicy.sigmoid(theta_val)
                    threshold_action_probability = MultiThresholdStoppingPolicy.stopping_probability(
                        b1=defender_stopping_probability, threshold=threshold, k=-20)
                    if s.id == 1:
                        stage_policy.append([1 - threshold_action_probability, threshold_action_probability])
                    elif s.id == 0:
                        stage_policy.append([threshold_action_probability, 1 - threshold_action_probability])
                    else:
                        raise ValueError(f"Invalid state: {s.id}, valid states are: 0 and 1")
                else:
                    stage_policy.append([0.5, 0.5])
            return stage_policy

    def _defender_action(self, o) -> Tuple[int, float]:
        """
        Multi-threshold stopping policy of the defender

        :param o: the input observation
        :return: the selected action (int) and its probability
        """
        b1 = o[1]
        l = int(o[0])
        if l > len(self.theta):
            l = len(self.theta)
        threshold = MultiThresholdStoppingPolicy.sigmoid(self.theta[l - 1])
        a, prob = MultiThresholdStoppingPolicy.smooth_threshold_action_selection(
            threshold=threshold, b1=b1, threshold_action=1, alternative_action=0)
        return a, prob

    @staticmethod
    def sigmoid(x) -> float:
        """
        The sigmoid function

        :param x: the input
        :return: sigmoid(x)
        """
        return 1 / (1 + math.exp(-x))

    @staticmethod
    def inverse_sigmoid(y) -> float:
        """
        The inverse sigmoid function

        :param y: sigmoid(x)
        :return: sigmoid(x)^(-1)
        """
        return math.log(y / (1 - y), math.e)

    @staticmethod
    def smooth_threshold_action_selection(threshold: float, b1: float, threshold_action: int = 1,
                                          alternative_action: int = 1, k=-20) -> Tuple[int, float]:
        """
        Selects the next action according to a smooth threshold function on the belief

        :param threshold: the threshold
        :param b1: the belief
        :param threshold_action: the action to select if the threshold is exceeded
        :param alternative_action: the alternative action to select if the threshold is not exceeded
        :return: the selected action and the probability
        """
        prob = MultiThresholdStoppingPolicy.stopping_probability(b1=b1, threshold=threshold, k=k)
        if random.uniform(0, 1) >= prob:
            return alternative_action, 1 - prob
        else:
            return threshold_action, prob

    @staticmethod
    def stopping_probability(b1, threshold, k=-20) -> float:
        """
        Returns the probability of stopping given a belief and a threshold

        :param b1: the belief
        :param threshold: the threshold
        :return: the stopping probability
        """
        if (1 - round(b1, 2)) == 0:
            return 1.0
        if round(b1, 2) == 0:
            return 0.0
        if (threshold * (1 - b1)) > 0 and (b1 * (1 - threshold)) / (threshold * (1 - b1)) > 0:
            try:
                return math.pow(1 + math.pow(((b1 * (1 - threshold)) / (threshold * (1 - b1))), k), -1)
            except Exception:
                return 0.0
        else:
            return 0.0

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the policy
        """
        d: Dict[str, Any] = {}
        d["theta"] = self.theta
        d["id"] = self.id
        d["simulation_name"] = self.simulation_name
        d["thresholds"] = self.thresholds()
        d["states"] = list(map(lambda x: x.to_dict(), self.states))
        d["actions"] = list(map(lambda x: x.to_dict(), self.actions))
        d["player_type"] = self.player_type
        d["agent_type"] = self.agent_type
        d["L"] = self.L
        if self.experiment_config is not None:
            d["experiment_config"] = self.experiment_config.to_dict()
        else:
            d["experiment_config"] = None
        d["avg_R"] = self.avg_R
        d["policy_type"] = self.policy_type
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "MultiThresholdStoppingPolicy":
        """
        Convert a dict representation of the object to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = MultiThresholdStoppingPolicy(
            theta=d["theta"], simulation_name=d["simulation_name"], L=d["L"],
            states=list(map(lambda x: State.from_dict(x), d["states"])), player_type=d["player_type"],
            actions=list(map(lambda x: Action.from_dict(x), d["actions"])),
            experiment_config=ExperimentConfig.from_dict(d["experiment_config"]), avg_R=d["avg_R"],
            agent_type=d["agent_type"])
        obj.id = d["id"]
        return obj

    def thresholds(self) -> List[float]:
        """
        :return: the thresholds
        """
        return list(map(lambda x: round(MultiThresholdStoppingPolicy.sigmoid(x), 3), self.theta))

    def stop_distributions(self) -> Dict[str, List[float]]:
        """
        :return: the stop distributions and their names
        """
        distributions = {}
        if self.player_type == PlayerType.DEFENDER:
            belief_space = np.linspace(0, 1, num=100)
            for l in range(1, self.L + 1):
                stop_dist = []
                for b in belief_space:
                    a1, prob = self._defender_action(o=[l, b])
                    if a1 == 1:
                        stop_dist.append(round(prob, 3))
                    else:
                        stop_dist.append(round(1 - prob, 3))
                distributions[constants.T_SPSA.STOP_DISTRIBUTION_DEFENDER + f"_l={l}"] = stop_dist
        else:
            defender_stop_space = np.linspace(0, 1, num=100)
            for s in self.states:
                if s.state_type != StateType.TERMINAL:
                    for l in range(1, self.L + 1):
                        stop_dist = []
                        for pi_1_S in defender_stop_space:
                            a2, prob = self._attacker_action(o=[l, pi_1_S, s.id], defender_stopping_prob=pi_1_S)
                            if a2 == 1:
                                stop_dist.append(round(prob, 3))
                            else:
                                stop_dist.append(round(1 - prob, 3))
                            distributions[constants.T_SPSA.STOP_DISTRIBUTION_ATTACKER + f"_l={l}_s={s.id}"] = \
                                stop_dist
        return distributions

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"theta: {self.theta}, id: {self.id}, simulation_name: {self.simulation_name}, " \
               f"thresholds: {self.thresholds()}, player_type: {self.player_type}, " \
               f"L:{self.L}, states: {self.states}, agent_type: {self.agent_type}, actions: {self.actions}," \
               f"experiment_config: {self.experiment_config}, avg_R: {self.avg_R}, policy_type: {self.policy_type}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "MultiThresholdStoppingPolicy":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return MultiThresholdStoppingPolicy.from_dict(json.loads(json_str))

    def copy(self) -> "MultiThresholdStoppingPolicy":
        """
        :return: a copy of the DTO
        """
        return self.from_dict(self.to_dict())



from typing import Dict, Any, List, Tuple
from csle_common.dao.emulation_config.emulation_env_state import EmulationEnvState
from csle_common.dao.system_identification.emulation_statistics import EmulationStatistics
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.defender.emulation_defender_action import EmulationDefenderAction
from csle_common.metastore.metastore_facade import MetastoreFacade
from csle_base.json_serializable import JSONSerializable


class EmulationStatisticsWindowed(JSONSerializable):
    """
    Windowed emulation statistic. The statistic is updated with the last <window size> samples
    """

    def __init__(self, window_size: int, emulation_name: str, descr: str):
        """
        Initializes the DTO

        :param window_size: the window size
        :param emulation_name: the emulation name
        :param descr: the description
        """
        self.window_size = window_size
        self.initial_states: List[EmulationEnvState] = []
        self.state_transitions: List[Tuple[EmulationEnvState, EmulationEnvState, EmulationDefenderAction,
                                           EmulationAttackerAction]] = []
        self.emulation_name = emulation_name
        self.descr = descr
        self.emulation_statistics = EmulationStatistics(emulation_name=self.emulation_name, descr=self.descr)
        self.statistics_id = MetastoreFacade.save_emulation_statistic(emulation_statistics=self.emulation_statistics)

    def add_initial_state(self, s: EmulationEnvState) -> None:
        """
        Adds an initial state sample

        :param s: the initial state
        :return: None
        """
        if len(self.initial_states) >= self.window_size:
            self.initial_states.pop(0)
        self.initial_states.append(s)

    def add_state_transition(self, s: EmulationEnvState, s_prime: EmulationEnvState, a1: EmulationDefenderAction,
                             a2: EmulationAttackerAction) -> None:
        """
        Adds a state transition samlpe

        :param s: the state
        :param s_prime: the next state
        :param a1: the defender action
        :param a2: the attacker action
        :return: None
        """
        if len(self.state_transitions) >= self.window_size:
            self.state_transitions.pop(0)
        self.state_transitions.append((s, s_prime, a1, a2))

    def update_emulation_statistics(self) -> None:
        """
        Updates the emulation statistic using the current window of samples

        :return: None
        """
        if len(self.state_transitions) == 0:
            return
        try:
            new_emulation_statistics = EmulationStatistics(emulation_name=self.emulation_name, descr=self.descr)
            print(f"updating statistic, num samples: {len(self.state_transitions)}")
            for i in range(len(self.state_transitions)):
                new_emulation_statistics.update_delta_statistics(
                    s=self.state_transitions[i][0], s_prime=self.state_transitions[i][1],
                    a1=self.state_transitions[i][2], a2=self.state_transitions[i][3])
            for i in range(len(self.initial_states)):
                new_emulation_statistics.update_initial_statistics(s=self.initial_states[i])
            self.emulation_statistics = new_emulation_statistics
            self.emulation_statistics.id = self.statistics_id
            MetastoreFacade.update_emulation_statistic(emulation_statistics=self.emulation_statistics,
                                                       id=self.statistics_id)
        except Exception:
            pass

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationStatisticsWindowed":
        """
        Converts a dict representation of the object into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = EmulationStatisticsWindowed(
            window_size=d["window_size"], emulation_name=d["emulation_name"], descr=d["descr"]
        )
        obj.emulation_statistics = EmulationStatistics.from_dict(d["emulation_statistics"])
        obj.statistics_id = d["statistics_id"]
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["window_size"] = self.window_size
        d["initial_states"] = list(map(lambda x: x.to_dict(), self.initial_states))
        d["state_transitions"] = list(map(lambda x: (x[0].to_dict(), x[1].to_dict(), x[2].to_dict(), x[3].to_dict()),
                                          self.state_transitions))
        d["emulation_name"] = self.emulation_name
        d["descr"] = self.descr
        d["emulation_statistics"] = self.emulation_statistics.to_dict()
        d["statistics_id"] = self.statistics_id
        return d

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationStatisticsWindowed":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationStatisticsWindowed.from_dict(json.loads(json_str))



from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_type import EmulationAttackerActionType
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_id import EmulationAttackerActionId
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_outcome import EmulationAttackerActionOutcome
from typing import List


class EmulationAttackerStoppingActions:
    """
    Class implementing stopping actions for the attacker in the emulation
    """

    @staticmethod
    def STOP(index: int) -> EmulationAttackerAction:
        """
        Reports a detected intrusion and stops

        :param index: index of the machine to apply the action to
        :return: the action
        """
        cmd: List[str] = []
        alt_cmd: List[str] = []
        return EmulationAttackerAction(id=EmulationAttackerActionId.STOP, name="Abort Intrusion", cmds=cmd,
                                       type=EmulationAttackerActionType.STOP,
                                       descr="Aborts an ongoing intrusion",
                                       index=index,
                                       ips=[], action_outcome=EmulationAttackerActionOutcome.GAME_END, alt_cmds=alt_cmd)

    @staticmethod
    def CONTINUE(index: int) -> EmulationAttackerAction:
        """
        A "continue" action, the attacker chooses to not make any action

        :param index: index of the machine to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.CONTINUE
        cmd: List[str] = []
        alt_cmd: List[str] = []
        return EmulationAttackerAction(id=id, name="Continue", cmds=cmd,
                                       type=EmulationAttackerActionType.CONTINUE,
                                       descr="A 'continue' action, the attacker chooses to not make any action",
                                       index=index,
                                       ips=[], action_outcome=EmulationAttackerActionOutcome.CONTINUE, alt_cmds=alt_cmd)
